/* TICKOS.CPP: OS/2 alarm clock server

   The Alarm Clock Server provides continuous and on-request time
   data in ASCIIZ format. The update frequency for continuous time
   display can be set by the client application. In addition, the
   server provides four distinctive alarm annunciators with client
   application controlled duration.

   The server id, as far as the client application is concerned, will
   always be the name of the file entered to execute the server,
   e.g., if the server file name is changed from TICKOS.EXE to
   CLOCK.EXE, a client application will need to identify the server
   as "CLOCK", not as "TICKOS".

   The "topic", "document" or "record" id for time data is "CLOCK".
   The server is NOT sensitive to the case of "clOck".

   The "item name" or "field" may be either "TIME" , "ALARM" or
   "INTERVAL". Once again, these are not case sensitive.

   Time display update frequency is supplied as a value which
   represents the time between updates. The frequency is changed
   via a WM_DDE_POKE where the value is an ASCIIZ number between
   1ooo (1 second) and 65,535 (many, many seconds).

   The alarms and their duration are implemented via a M_DDE_EXECUTE
   message. An alarm command takes the form:

                    "ALARMn,xxxxxx"

   where 'n' is a number between 1 and 4 and 'xxxxxx' is a number
   between 0 and 999999. 'n' selects one of the four alarms and
   'xxxxxx' sets the duration of the alarm.

   The server side of the demo is very much like the application
   side and the user should consdier many of the descriptions for
   TICKOSC as commentary on the server.

   In addition to being a demonstration of DDE, the Alarm Clock
   server also demonstrates:

   - bidirectional communication between a client and object window
   - use of a listbox and listbox item handles
   - use of the bytes in a window reserved for an application's use
     (QWL_USER) to support an instance data area (IDA)

 ====================================================================
   NOTES:
           (1) TICKOS requires the following files:
                 TICKOS.CPP  - TICKOS main(), window proc & functions
                 TICKOS.H    - TICKOS constants, structures & prototypes
                 TICKOS.RC   - TICKOS resources
                 TICKOSS.ICO - icon
           (2) In the comments I have used the term 'sender'
               rather than 'client' to avoid confusion between
               the server's client window and the client
               application.
           (3) In the comments I refer to a data packet; this is
               always a reference to the DDESTRUCT instance which is
               used as a vehicle for passing messages between the
               sender/server and server/sender.
           (4) The extensive commenting may be considered a bit
               too much by some; however, I intended this source
               to be a working compendium of all the information
               that is spread over at least two manuals in the
               IBM documentation and to provide an alternative
               phrasing of the text of these manuals.
           (5) The IBM documentation (both online and hardcopy)
               consistently refers to the DDESTRUCT data member
                "usStatus". This is an error; the member is actually
                "fsStatus".
           (6) This server doesn't have any use for WM_DDE_DATA or
               WM_DDE_INITIATEACK messages. However, it is prepared
               to deal with the unexpected receipt of these messages
               by doing the most basic message housekeeping and
               reporting the receipt to the client. Consequently,
               the WM_DDE_DATA and WM_DDE_INITIATEACK handlers
               represent the barebones of the two types of DDE message
               handlers.

  -------------------------------------------------------------------
     PM FUNCTIONS DEMONSTRATED:
       WinInitialize()              WinCreateMsgQueue()
       WinRegisterClass()           WinCreateWindow()
       WinMessageBox()              WinDestroyMsgQueue()
       WinTerminate()               WinQueryWindowRect()
       WinSetWindowPos()            WinGetMsg()
       WinDispatchMsg()             WinDestroyWindow()
       WinQueryWindow()             WinWindowFromID()
       WinSendMsg()                 WinGetPS()
       GpiQueryFontMetrics()        WinReleasePS()
       WinSetWindowPtr()            WinQueryWindowPtr()
       WinBeginPaint()              WinFillRect()
       WinDrawText()                WinEndPaint()
       WinInvalidateRect()          WinAlarm()
       WinPostMsg()                 WinDefWindowProc()
       WinDdeRespond()              WinDdePostMsg()
       DosFreeMem()                 WinStartTimer()
       WinStopTimer()               WinQueryAnchorBlock()
       DosAllocSharedMem()          WinQueryWindowProcess()
       DosGiveSharedMem()           DosGetDateTime()
       DosBeep()                    DosSleep()  

  -------------------------------------------------------------------
               Copyright (c) 1993 Borland International
  -------------------------------------------------------------------
     04-04-93 EAS   : i hate initial releases
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <dir.h>
#include <process.h>

#define INCL_WIN
#define INCL_GPI
#define INCL_DOS
#include <os2.h>

#include "tickos.h"

USHORT CLIENTIDA::usInstance = 0;   // link counter initialization

SHORT wDateTime[6];
SHORT DaysPerMonth[13] = {0,31,28,31,30,31,30,31,31,30,31,30,31};

PCHAR  szCmd[] = { "ALARM1", "ALARM2", "ALARM3", "ALARM4" };
USHORT commands = 4;

PCHAR erArray[] = { "not responding",    // APPNOTRESPONDING
                    "busy",              // APPBUSY
                    "rejected data" };   // APPREJECTEDDATA

// Note that the comments on this array may be only be true for
// OS/2 2.0. They are supplied here as a convenience (and, because
// I just cut-and-pasted-and-edited the data from PMWIN.H it seemed
// like a waste to delete this info). Since WM_DDE_FIRST is also
// defined and I'll bet that it will always be guaranteed to be
// the value of the first (i.e., WM_DDE_INITIATE) DDE message, as
// long as you use this value you should NOT (nor need to) hardwire
// these values into your code.
PCHAR DDEmsgArray[] = { "WM_DDE_INITIATE",      //0x00A0
                        "WM_DDE_REQUEST",       // 0x00A1
                        "WM_DDE_ACK",           // 0x00A2
                        "WM_DDE_DATA",          // 0x00A3
                        "WM_DDE_ADVISE",        // 0x00A4
                        "WM_DDE_UNADVISE",      // 0x00A5
                        "WM_DDE_POKE",          // 0x00A6
                        "WM_DDE_EXECUTE",       // 0x00A7
                        "WM_DDE_TERMINATE",     // 0x00A8
                        "WM_DDE_INITIATEACK" }; // 0x00A9


//*******************************************************************
#pragma argsused
INT main(USHORT arg, PCHAR argv[] ) {
  CHAR szAppName[_MAX_FNAME];

  // Parse path to get executing file name.
  _splitpath(argv[0], NULL, NULL, szAppName, NULL);
  PCHAR pszAppName = new char[strlen(szAppName)+1];
  strcpy(pszAppName, szAppName);

  // Setup message queue.
  HAB hab = WinInitialize(0);
  HMQ hmq = WinCreateMsgQueue(hab, 0);

  // Register the server client and object window classes.
  WinRegisterClass(hab, (PSZ)szClientClass, clientWndProc,
                   CS_CLIPCHILDREN|CS_SIZEREDRAW, sizeof(PCLIENTIDA));

  WinRegisterClass(hab, (PSZ)szObjClass, objWndProc,
                   CS_CLIPCHILDREN|CS_SIZEREDRAW, sizeof(POBJIDA));

  // Setup frame control data and...
  FRAMECDATA fcdata;
  fcdata.cb = sizeof(FRAMECDATA);
  fcdata.flCreateFlags = FCF_TITLEBAR      |
                         FCF_SYSMENU       |
                         FCF_SIZEBORDER    |
                         FCF_MINMAX        |
                         FCF_SHELLPOSITION |
                         FCF_TASKLIST      |
                         FCF_MENU          |
                         FCF_ICON;

  fcdata.hmodResources =(HMODULE)0L;
  fcdata.idResources   = idFrameWindow;

  // ...create server's frame and client windows.
  HWND hwndFrame = WinCreateWindow(HWND_DESKTOP, WC_FRAME,
                                   (PSZ)0L, 0, 0,0,0,0,
                                   (HWND)0L, HWND_TOP, idFrameWindow,
                                   &fcdata, (PVOID)0L);

  if (!hwndFrame) {
    CHAR errbuf[100];
    sprintf(errbuf, "%s frame window creation error %X",
            szAppName,
            WinGetLastError(WinQueryAnchorBlock(HWND_DESKTOP)));
    WinMessageBox(HWND_DESKTOP, HWND_DESKTOP, (PSZ)errbuf,
                  (PSZ)"Error", 0, MB_ERROR|MB_ENTER);
    WinDestroyMsgQueue(hmq);
    WinTerminate(hab);
    return 1;
  }

  // Application (EXE file name) is passed to WM_CREATE so it
  // can be inserted into the client window's instance data area.
  HWND hwndClient = WinCreateWindow(hwndFrame, (PSZ)szClientClass,
                                    (PSZ)0L, 0,
                                    0,0,0,0, (HWND)0L, HWND_TOP,
                                    FID_CLIENT, (PVOID)pszAppName, 0L);
  if (!hwndClient) {
    CHAR errbuf[100];
    sprintf(errbuf, "%s client window creation error %X",
            szAppName,
            WinGetLastError(WinQueryAnchorBlock(HWND_DESKTOP)));
    WinMessageBox(HWND_DESKTOP, HWND_DESKTOP, (PSZ)errbuf,
                  (PSZ)"Error", 0, MB_ERROR|MB_ENTER);
    WinDestroyWindow(hwndFrame);
    WinDestroyMsgQueue(hmq);
    WinTerminate(hab);
    return 1;
  }

  RECTL rcl;
  WinQueryWindowRect(HWND_DESKTOP, &rcl);
  USHORT yHite = USHORT(rcl.yTop/4);
  USHORT xWidth = USHORT(rcl.xRight-rcl.xRight/8);
  USHORT cx = USHORT(rcl.xLeft+5);
  USHORT cy = USHORT(rcl.yTop-yHite-5);
  WinSetWindowPos(hwndFrame, HWND_TOP, cx, cy, xWidth, yHite,
                  SWP_SIZE|SWP_MOVE|SWP_SHOW);

  // Start message loop processing.
  QMSG qmsg;
  while(WinGetMsg(hab, &qmsg, 0L, 0, 0))
    WinDispatchMsg(hab, &qmsg);

  // We are here because message loop processing encountered a
  // WM_QUIT; dump windows and queue, terminate process and
  // exit.
  WinDestroyWindow(hwndFrame);
  WinDestroyMsgQueue(hmq);
  WinTerminate(hab);
  return 0;
} // end of main()


/*-------------------------------------------------------------------
    clientWndProc - handles all user input, initializes the DDE link
                    upon receipt of a WM_DDE_INITIATE message and
                    displays link status in client window.
*/

MRESULT EXPENTRY clientWndProc(HWND hwnd, ULONG msg,
                               MPARAM mp1, MPARAM mp2) {
  MRESULT            mr = FALSE;
  HPS                hps;
  RECTL              rcl;
  PCLIENTIDA         pClientIDA;
  POBJIDA            pObjIDA;
  SHORT              cy;

  switch(msg) {

    //
    case WM_CREATE:
      {
        // Create server's client window IDA.
        pClientIDA = new CLIENTIDA(hwnd);

        // Uncheck report-all-errors option
        HWND hwndFrame = WinQueryWindow(hwnd, QW_PARENT);
        HWND hwndMenu = WinWindowFromID(hwndFrame, FID_MENU);
        WinSendMsg(hwndMenu, MM_SETITEMATTR,
                   MPFROM2SHORT(IDM_ALLERRORS, TRUE),
                   MPFROM2SHORT(MIA_CHECKED, ~MIA_CHECKED));

        // Assure there's something in the text slot and load the
        // application name.
        strcpy(pClientIDA->szText, " ");
        pClientIDA->pszAppName = (PCHAR)PVOIDFROMMP(mp1);

        // Get font metrics.
        hps = WinGetPS(hwnd);
      	GpiQueryFontMetrics(hps, sizeof(pClientIDA->fontMetrics),
                            &pClientIDA->fontMetrics);
        WinReleasePS(hps);

        // LS_NOADJUSTPOS permits the listbox to always fill the
        // the calculated rectangle exactly; without it there will
        // will be a varying amount of white space at the top of the
        // listbox for many vertical frame sizes.

        // 'fontMetrics.lMaxBaselineExt' is the height of a line of
        // text in the current font and is used to offset the listbox
        // one text line height from the bottom of the window frame.
        // The exposed client window area is used to display server
        // status messages, i.e., error messages.
        WinQueryWindowRect(hwnd, &rcl);
        cy = (SHORT)pClientIDA->fontMetrics.lMaxBaselineExt;
        pClientIDA->hwndListbox = WinCreateWindow(hwnd, WC_LISTBOX,
                                0L, WS_VISIBLE|LS_NOADJUSTPOS, 0,
                                cy, rcl.xRight, rcl.yTop-cy, hwnd,
                                HWND_TOP, IDLB_LISTBOX, 0L, 0L);

        // Set foreground and background colors. (These are the
        // colors of the exposed client area. The listbox has its
        // own colors.)
        pClientIDA->foreground = CLR_YELLOW;
        pClientIDA->background = CLR_RED;

        // Set a pointer to IDA into the client window user slot.
        WinSetWindowPtr(hwnd, QWL_USER, pClientIDA);
      }
      break;

    //
    case WM_PAINT:

      // Update status line with text currently in 'szText' data
      // member of client window's IDA. Given this paint case, all
      // that is required to update the "status line" is to:
      //   - get the client's IDA
      //   - strcpy() or sprintf() a message into its 'szText'
      //   - call WinInvalidateRect()
      pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      hps = WinBeginPaint(hwnd, 0L, &rcl);
      WinFillRect(hps, &rcl, pClientIDA->background);
      rcl.xLeft = 0;
      rcl.yTop = pClientIDA->fontMetrics.lMaxBaselineExt+1;
      WinDrawText(hps, -1, (PSZ)pClientIDA->szText, &rcl,
                  pClientIDA->foreground, pClientIDA->background,
                  DT_VCENTER);
      WinEndPaint(hps);
      break;

    //
    case WM_SIZE:
      pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);

      // (Re)size the list box.
      if (pClientIDA->hwndListbox != 0L) {
        cy = (SHORT)pClientIDA->fontMetrics.lMaxBaselineExt;
        WinSetWindowPos(pClientIDA->hwndListbox, HWND_TOP, 0,
                        cy, SHORT1FROMMP(mp2),
                        SHORT2FROMMP(mp2)-cy,
                        SWP_MOVE|SWP_SIZE|SWP_SHOW);
        return FALSE;
      }
      break;

    //
    // A sender application wants to initiate a DDE link.
    // WinDdeInitiate() SENDS this message to all top level windows.
    // The server responds by sending back a WM_DDE_INITIATEACK
    //   mp1 = sender's window handle
    //   mp2 = pointer to DDEINIT instance
    // --------------------------------------------------------
    //   typedef struct _DDEINIT {
    //      ULONG cb;          // length of structure
    //      PSZ   pszAppName;  // application name
    //      PSZ   pszTopic;    // topic
    //      ULONG offConvtext; // conversation context
    //   } DDEINIT;
    // --------------------------------------------------------
    //  pszAppName - if 0 length, any application can respond
    //  pszTopic   - if 0 length, each responding applicatiom
    //               responds once for each topic it can support.
    case WM_DDE_INITIATE:
      {
        // Get the sender's window handle and initialization packet.
        HWND hwndSender = HWNDFROMMP(mp1);
        PDDEINIT pDdeInit = (PDDEINIT)PVOIDFROMMP(mp2);

        // If we can handle the requested topic, let's setup the
        // resources needed for the link.
        pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
        if ((!strcmpi((PCHAR)pDdeInit->pszAppName,
                      (PCHAR)pClientIDA->pszAppName)) &
            (!strcmpi((PCHAR)pDdeInit->pszTopic, (PCHAR)szTopic))) {

          // Create the link object window and its IDA (instance data
          // area), initialize the IDA and send a WM_DDE_INITIATEACK
          // to the application that originated the WM_DDE_INITIATE.
          // The object window's window procedure subsequently
          // processes all DDE exchanges between the Alarm Clock server
          // and the linked application and communicates relevant infor-
          // mation regarding these exchanges to this client window
          // via PM messages. Each link created gets its own object
          // window.
          makeLink(hwnd, hwndSender);
        }
      }
      return (MRESULT)TRUE;

    //---------------------------------------------------------------
    // An object window says an app has established a link with us.
    // 'mp1' contains the object window's IDA.
    case UM_CLIENTLINKED:
      {
        // Insert new link into list.

        // Get client/object window IDAs, increment the link
        // counter and update the status line buffer.
        pObjIDA = (POBJIDA)PVOIDFROMMP(mp1);
        pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
        ++pClientIDA->usInstance;
        sprintf(pClientIDA->szText, " %d DDE apps linked",
                pClientIDA->usInstance);

        // Use sender's window handle as ID when the new client
        // link is added to the list.
        CHAR szText[80];
        sprintf(szText, "(%p) > linked", pObjIDA->hwndSender);
        SHORT index = (SHORT)WinSendMsg(pObjIDA->hwndListbox,
                                        LM_INSERTITEM,
                                        MPFROMSHORT(LIT_END),
                                        MPFROMP(szText));

        // Now associate the link's object window IDA with
        // this entry and...
        WinSendMsg(pObjIDA->hwndListbox, LM_SETITEMHANDLE,
                   MPFROMSHORT(index), MPFROMP(pObjIDA));

        // ...make sure new entry appears in listbox and...
        WinSendMsg(pObjIDA->hwndListbox, LM_SETTOPINDEX,
                   MPFROMSHORT(index), 0L);

        // ...force update of status line.

        WinInvalidateRect(hwnd, 0L, 0L);
      }
      break;

    //---------------------------------------------------------------
    // Client window has received a link message from the object
    // window.
    //   mp1 = object window IDA pointer
    //   mp2 = message id
    case UM_LINKMESSAGE:
      pObjIDA = (POBJIDA)PVOIDFROMMP(mp1);
      pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);

      switch(SHORT1FROMMP(mp2)) {
        case WM_DDE_REQUEST:
          sprintf(pObjIDA->szText, "WM_DDE_REQUEST: Single shot"
                                   " update");
          break;
        case WM_DDE_ADVISE:
          sprintf(pObjIDA->szText, "WM_DDE_ADVISE: Start continuous"
                                   " update");
          break;
        case WM_DDE_UNADVISE:
          sprintf(pObjIDA->szText, "WM_DDE_UNADVISE: Stop"
                                   " continuous update");
          break;
        case WM_DDE_POKE:
          sprintf(pObjIDA->szText, "WM_DDE_POKE: Interval"
                                   " changed to %d",
                  pObjIDA->interval);
          break;
        case WM_DDE_EXECUTE:
          sprintf(pObjIDA->szText, "WM_DDE_EXECUTE: ALARM %d "
                                   " for %d seconds",
                  pObjIDA->usAlarm+1,
                  pObjIDA->usDuration/1000+1);
          break;

        // On a WM_DDE_TERMINATE we find and delete listbox entry
        // for terminated link.
        case WM_DDE_TERMINATE:
          {
            // We need both the object window and client window IDA.
            pObjIDA = (POBJIDA)PVOIDFROMMP(mp1);
            pClientIDA =
                  (PCLIENTIDA)WinQueryWindowPtr(pObjIDA->hwndClient,
                                                QWL_USER);

            // Get number of entries in list and...
            SHORT count = (SHORT)WinSendMsg(pClientIDA->hwndListbox,
                                            LM_QUERYITEMCOUNT, 0L, 0L);

            // for each of them...
            POBJIDA temp;
            for (USHORT i=0; i<count; i++) {

              // ...get associated object window IDA and...
              temp = (POBJIDA)WinSendMsg(pClientIDA->hwndListbox,
                                         LM_QUERYITEMHANDLE,
                                         MPFROMSHORT(i), 0L);

              // ...see if it has the same sender as the one which
              // dropped the link. If so, delete the matching list
              // entry and exit else loop until we find a match or
              // have exhausted entries.
              if (temp->hwndSender == pObjIDA->hwndSender) {
                WinSendMsg(pClientIDA->hwndListbox, LM_DELETEITEM,
                           MPFROMSHORT(i), 0L);

                // Decrement the link counter, update status line
                // information and...
                if (--pClientIDA->usInstance)
                  sprintf(pClientIDA->szText, " %d DDE apps linked",
                          pClientIDA->usInstance);
                else
                  *pClientIDA->szText = '\0';

                // ...force update of status line.
                WinInvalidateRect(hwnd, 0L, 0L);

                // Since the list box entry is history skip the
                // attempt to update it.
                return FALSE;
              }
            }
          }
          break;
        case WM_DDE_DATA:

          // Alarm clock server does nothing with WM_DDE_DATA
          // and shouldn't receive one.
          sprintf(pObjIDA->szText, "WM_DDE_DATA: What????");
          break;
      }
      setListboxText(pObjIDA);
      break;

    //---------------------------------------------------------------
    // Object window reports a link error:
    //   mp1 = object window IDA pointer
    //       = 0 if this is a failure-to-open-object-window error
    //   mp2:  low byte = DDE message ID    (SHORT1FROMMP)
    //         hi byte  = error ID          (SHORT2FROMMP)
    case UM_LINKERROR:
      {
        pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
        pObjIDA = (POBJIDA)PVOIDFROMMP(mp1);
        USHORT erm = SHORT2FROMMP(mp2);  // error ID
        USHORT eri = SHORT1FROMMP(mp2);  // DDE message ID

        if (!pObjIDA) {
          strcpy(pClientIDA->szText,
                 " OBJECT WINDOW IDA CREATION ERROR:"
                 " Link failure at server.");
        }
        else {

          // Note that WM_DDE_FIRST is a standard define which makes
          // it easy (and revision independent) to make the WM_DDE
          // messages 0-relative for array use (as is demonstrated
          // here.
          sprintf(pClientIDA->szText, " (%p) error: %s on %s",
                  pClientIDA->usInstance,
                  erArray[eri],
                  DDEmsgArray[erm-WM_DDE_FIRST]);
        }

        // ...force update of status line.
        WinInvalidateRect(hwnd, 0L, 0L);
      }
      break;

    //
    // Server menu item selected.
    case WM_COMMAND:
      switch(SHORT1FROMMP(mp1)) {

        case IDM_EXIT:
          WinPostMsg(hwnd, WM_CLOSE, 0L, 0L);
          break;

        // Toggle report-all-errors option.
        case IDM_ALLERRORS:
          {
            HWND hwndFrame = WinQueryWindow(hwnd, QW_PARENT);
            HWND hwndMenu = WinWindowFromID(hwndFrame, FID_MENU);
            pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
            if (toggleCheck(hwndMenu, IDM_ALLERRORS)) {
              pClientIDA->usFlags |= FALLERRORS;
              if (pClientIDA->hwndObj)
                WinPostMsg(pClientIDA->hwndObj, UM_ALLERRORS,
                           MPFROMSHORT(TRUE), 0L);
            }
            else {
              pClientIDA->usFlags &= ~FALLERRORS;
              if (pClientIDA->hwndObj)
                WinPostMsg(pClientIDA->hwndObj, UM_ALLERRORS,
                           MPFROMSHORT(FALSE), 0L);
            }
          }
          break;

        case IDM_ABOUT:
          WinMessageBox(HWND_DESKTOP, hwnd,
                        (PSZ)"BC++ for OS/2 DDE demo",
                        (PSZ)"Alarm Clock Server",
                        1, MB_OK|MB_APPLMODAL);
          break;

        default:
          mr = WinDefWindowProc(hwnd, msg, mp1, mp2);
          break;
      }
      break;

    //
    // Just for the heck of it we'll shut down a sender's DDE link
    // if the operator double clicks (or presses enter) on the link's
    // listbox entry. I seriously doubt that a server should allow
    // this kind of operation but I think it demonstrates something
    // worthwhile.
    case WM_CONTROL:
      switch(SHORT1FROMMP(mp1)) {

        // Heard from the listbox lately? Yah... Wha'd it say?
        case IDLB_LISTBOX:
          switch(SHORT2FROMMP(mp1)) {

            // Someone's hit the entry key or double clicked on
            // one of my entries.
            case LN_ENTER:
              {

                // In that case (pun intended) let's get the listbox
                // handle and...
                HWND hwndListbox = HWNDFROMMP(mp2);

                // ...see which of its entries is being messed  with.
                USHORT item;
                item = (USHORT)WinSendMsg(hwndListbox,
                                          LM_QUERYSELECTION,
                                          0L, 0L);

                // Get the entry's baggage and...
                POBJIDA temp;
                temp = (POBJIDA)WinSendMsg(hwndListbox,
                                           LM_QUERYITEMHANDLE,
                                           MPFROMSHORT(item), 0L);

                // ...use some of the contents to terminate the
                // entry's link to us.
                if (WinPostMsg(temp->hwnd, WM_DDE_TERMINATE,
                               MPFROMHWND(temp->hwndSender), 0L)) {
                  WinSendMsg(hwndListbox, LM_DELETEITEM,
                             MPFROMSHORT(item), 0L);

                  // Decrement the link counter, update status line
                  // information and...
                  if (--pClientIDA->usInstance)
                    sprintf(pClientIDA->szText, " %d DDE apps linked",
                            pClientIDA->usInstance);
                }
              }
              break;
          }
          break;
      }
      break;

    //
    // Shut down server.
    case WM_CLOSE:
      pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      delete pClientIDA->pszAppName;
      delete pClientIDA;
      WinPostMsg(hwnd, WM_QUIT, 0L, 0L);
     break;

   default:
     mr = WinDefWindowProc(hwnd, msg, mp1, mp2);
     break;
  }

  return mr;
} // end of clientWndProc()


/*-------------------------------------------------------------------
   makeLink - sets up a DDE link by creating an Object Window to
              maintain a DDE exchange. It allocates and initializes
              private data for the Object Window to use.
*/

VOID makeLink(HWND hwndClient, HWND hwndSender) {

  // Load application name from server's client window IDA.
  PCLIENTIDA pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwndClient,
                                                        QWL_USER);

  // Allocate object window instance data area and initialize
  // link specific data members.
  POBJIDA pObjIDA = new OBJIDA(hwndClient, hwndSender,
                               pClientIDA->pszAppName,
                               szTopic);

  // If IDA allocation failed we might as well give it up!
  // Report condition to client window and get out'ta here.
  if (!pObjIDA) {
    WinSendMsg(hwndClient, UM_LINKERROR, 0L, 0L);
    return;
  }

  // If all-errors flag is set, enable all error reporting.
  if (pClientIDA->usFlags&FALLERRORS)
    pObjIDA->bAllErrors = TRUE;

  // Load default item, listbox handle and set default display
  // interval for once/second.
  pObjIDA->pszItemName = new char[strlen(szItemName)+1];
  strcpy(pObjIDA->pszItemName, szItemName);
  pObjIDA->hwndListbox = pClientIDA->hwndListbox;
  pObjIDA->interval = 1000;

  // Create the object window; objWndProc() will receive the
  // 'pObjIDA' pointer in 'mp1' of its WM_CREATE message.
  WinCreateWindow(HWND_OBJECT, (PSZ)szObjClass,
                  NULL, 0L, 0, 0, 0, 0,
                  hwndClient, HWND_TOP, 0L,
                  pObjIDA, 0L);
} // end of makeLink()


/*-------------------------------------------------------------------
    objWndProc - handles the following DDE messages:

    WM_DDE_ACK:         if DDE_FACKREQ bit is set in sender's service
                        request data packet, server waits for this
                        message from the sender before sending updated
                        data; is also responsible for request data
                        packet if DDE_FACKREQ is set.
    WM_DDE_REQUEST:     sends time to application once
    WM_DDE_POKE:        changes the update interval
    WM_DDE_ADVISE:      sends time once per update interval
    WM_DDE_UNADVISE:    stops sending time once per update interval
    WM_DDE_EXECUTE:     executes one of four alarm annunciators
    WM_DDE_TERMINATE:   Server frees its DDE link object window, and
                        confirms termination to sender.
    WM_DDE_INITIATEACK:
    WM_DDE_DATA:        We don't do anything with WM_DDE_DATA except
                        notify the client window of its receipt.

    NOTE: This procedure handles all DDE link messages EXCEPT a
          WM_DDE_INITIATE which must be handled by the client window
          procedure.
*/

MRESULT EXPENTRY objWndProc(HWND hwnd, ULONG msg,
                            MPARAM mp1, MPARAM mp2) {
  MRESULT    mr = FALSE;
  PCLIENTIDA pClientIDA;
  POBJIDA    pObjIDA;
  HWND       hwndSender;
  PDDESTRUCT pDdeStruct;
  HAB        habServer;

  switch(msg) {

    //
    // Setup initial state of the server object window.
    case WM_CREATE:

      // The WinCreateWindow() passes this message a partially
      // initialized instance data area (IDA) whose initialization
      // is topped off and a pointer to it is stuffed into the slot
      // in all windows (QWL_USER) reserved for an application's use.
      pObjIDA = (POBJIDA)PVOIDFROMMP(mp1);
      WinSetWindowPtr(hwnd, QWL_USER, pObjIDA);

      // Load object window's handle, get client IDA and load client
      // window 's handle.
      pObjIDA->hwnd = hwnd;
      pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(pObjIDA->hwndClient,
                                                 QWL_USER);
      pClientIDA->hwndObj = hwnd;

      // Let the sender know that our end of the DDE link is
      // anchored and...
      WinDdeRespond(pObjIDA->hwndSender, hwnd,
                    (PSZ)pObjIDA->pszAppName,
                    (PSZ)pObjIDA->pszTopic, NULL);

      // ...let client window know we are linked to a new sender.
      // We pass the object window's IDA back to the client window
      // in 'mp1' so that it can do what it will based on this
      // event.
      WinPostMsg(pObjIDA->hwndClient, UM_CLIENTLINKED,
                 MPFROMP(pObjIDA), 0L);
      break;  // end of WM_CREATE

    case UM_ALLERRORS:
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      if (mp1)
        pObjIDA->bAllErrors = TRUE;
      else
        pObjIDA->bAllErrors = FALSE;
      break;

    //
    // Data update requested.
    //   mp1 = server's window handle
    //   mp2 = pointer to DDESTRUCT instance
    //      offszItemName = item name to get
    //      usFormat      = DDE format in which data is to be
    //                      returned.
    case WM_DDE_REQUEST:

      // WM_DDE_REQUEST is a request for oneshot data update.
      // However, if the DDE_FNODATA bit is set in the data packet
      // passed to WM_DDE_ADVISE, WM_DDE_ADVISE processing will
      // transmit a DDE packet without data to signal that a data
      // update is available and the sender is expected to get
      // the data via an WM_DDE_REQUEST.

      // Get sender's window handle, the DDE data packet and
      // and the server's IDA.
      hwndSender = HWNDFROMMP(mp1);

      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);

      // If this is the right server/sender match...
      if (pObjIDA->hwndSender == hwndSender) {

        // ...and this server object window is handling the item
        // requested by the sender...
        if (!strcmpi((PCHAR)DDES_PSZITEMNAME(pDdeStruct),
                     (PCHAR)szItemName)) {

          // Preserve the ack-request flag and release the sender's
          // data packet.
          USHORT facreq = USHORT(pDdeStruct->fsStatus&DDE_FACKREQ);
          DosFreeMem(pDdeStruct);

          // Get time and plug it into a newly created data packet.
          // Note that state of sender's FACKREQ bit has been
          // preserved in this packet.
          PCHAR szTime = getTime();
          pDdeStruct = createDataPacket(hwndSender,
                                        (PSZ)pObjIDA->pszItemName,
                                        facreq, (PSZ)szTime,
                                        USHORT(strlen(szTime)+1));

          // Client app may be requesting data after receiving a
          // 0-length data packet (NODATA mode); flag this as a
          // WM_DDE_ADVISE response else...
          if (pObjIDA->bAdvise)
            pDdeStruct->fsStatus &= ~DDE_FRESPONSE;
          else

            // ...this is REALLY a response to a WM_DDE_REQUEST
            // message.
            pDdeStruct->fsStatus |= DDE_FRESPONSE;

          // At this point the data packet is complete and:
          //  DDESTRUCT->fsStatus:
          //     DDE_FACREQ =    If set client receiver should
          //                     return an WM_DDE_ACK message
          //     DDE_FRESPONSE = If set data is a reponse to a
          //                     WM_DDE_REQUEST else it is a
          //                     response to a WM_DDE_ADVISE

          // Ship the packet off to the sender, deleting packet
          // if the sender is hard-of-hearing or acknowledgment
          // isn't required.
          if (!WinDdePostMsg(hwndSender, hwnd, WM_DDE_DATA,
                             pDdeStruct, TRUE) || !facreq) {
            DosFreeMem(pDdeStruct);

            // If acknowledge-request is in effect but we are
            // here anyway it must be because the sender isn't
            // listening!!!
            if (facreq)
              WinPostMsg(pObjIDA->hwndClient, UM_LINKERROR,
                         MPFROMP(pObjIDA),
                         MPFROM2SHORT(APPNOTRESPONDING,
                                      WM_DDE_REQUEST));
          }

          // Report action to client window.
          WinPostMsg(pObjIDA->hwndClient, UM_LINKMESSAGE,
                     MPFROMP(pObjIDA), MPFROMSHORT(WM_DDE_REQUEST));
        }
      }
      break;  // end of WM_DDE_REQUEST

    //
    // Continuous data update requested.
    //   mp1 = sender's window handle
    //   mp2 = pointer to DDESTRUCT instance
    case WM_DDE_ADVISE:

      // Nothing is actually shipped from WM_DDE_ADVISE; all
      // transmissions happen in WM_TIMER and, if DDE_FNODATA is
      // enabled, from WM_DDE_REQUEST.

      // Get the sender's window handle, pointer to DDESTRUCT
      // instance passed by sender and server object window IDA.
      hwndSender = HWNDFROMMP(mp1);
      pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

      // The advise flag is set here and reset by an UNADVISE.
      // This is necessary since a WM_DDE_REQUEST may, in fact,
      // be the message by which the sender ultimately gets an
      // WM_DDE_ADVISE sponsored message. This flag is used by
      // WM_DDE_RESPONSE processing to determine the originating
      // state of the 'DDE_FRESPONSE' fsStatus bit of a
      // WM_DDE_DATA data packet.
      pObjIDA->bAdvise = TRUE;

      // If this is the right server for the sender...
      if (pObjIDA->hwndSender == hwndSender) {

        // ..and we are serving up the target item (case is ignored)...
        if (!strcmpi((PCHAR)DDES_PSZITEMNAME(pDdeStruct),
                     (PCHAR)szItemName)) {

          // ...and sender wants the server to delay shipment of
          // data until it receives an WM_DDE_ACK from sender...
          if (pDdeStruct->fsStatus&DDE_FACKREQ)

            // ...store fact in server's IDA.
            pObjIDA->bAcknowledgeRequest = TRUE;

          // Ditto if the server wants WM_DDE_ADVISE processing
          // to indicate that updated data is available rather than
          // actually send the data; sender will get data via a
          // WM_DDE_REQUEST.
          if (pDdeStruct->fsStatus&DDE_FNODATA)
            pObjIDA->bNotifyUpdate = TRUE;

          // Now start the timer and...
          habServer = WinQueryAnchorBlock(hwnd);
          WinStartTimer(habServer, hwnd, pObjIDA->usTimerNo,
                        pObjIDA->interval);

          // ...report action to client window.
          WinPostMsg(pObjIDA->hwndClient, UM_LINKMESSAGE,
                     MPFROMP(pObjIDA), MPFROMSHORT(WM_DDE_ADVISE));
        }
      }

      // Release the sender's data packet.
      DosFreeMem(pDdeStruct);
      break;  // end of WM_DDE_ADVISE

    //
    // Termination of ADVISE data transfer requested.
    // The DDESTRUCT data packet may specify the item name of the
    // data transfer that is to be discontinued. If it doesn't
    // (offszItemName is the offset to a zero length string) the
    // sender is requesting that we terminate all item links being
    // supported by this server.
    //   mp1 = sender's window handle
    //   mp2 = DDESTRUCT instance pointer
    case WM_DDE_UNADVISE:
      {
        // Get sender's window handle, data packet, and the
        // server's IDA.
        hwndSender = HWNDFROMMP(mp1);
        pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);
        pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

        // Flag us out of advise mode so subsequent WM_DDE_REQUEST
        // data packets are flagged as responses to a WM_DDE_REQUEST.
        pObjIDA->bAdvise = FALSE;

        // Server/sender match? If so...
        if (pObjIDA->hwndSender == hwndSender) {

          // ...transfer the item name to a local buffer so
          // we can dump the data packet.
          PCHAR itemName = (PCHAR)DDES_PSZITEMNAME(pDdeStruct);
          PCHAR itembuf = new char[strlen(itemName)+1];
          strcpy(itembuf, itemName);

          // We got the right item? If so...
          if (!strcmpi(itembuf, szItemName)) {

            // ...stop the timer (which effectively stops ADVISE
            // data update).
            habServer = WinQueryAnchorBlock(hwnd);
            WinStopTimer(habServer, hwnd, pObjIDA->usTimerNo);

            // Flag a positive ackowledge in data packet and
            // reset ack-on-receipt flag.
            pDdeStruct->fsStatus |= DDE_FACK;
            pObjIDA->bAcknowledgeRequest = TRUE;

            // Report action to client window.
            WinPostMsg(pObjIDA->hwndClient, UM_LINKMESSAGE,
                       MPFROMP(pObjIDA), MPFROMSHORT(WM_DDE_UNADVISE));
          }
          else

            // ...do nothing but flag a negative acknowledgment.
            pDdeStruct->fsStatus &= ~DDE_FACK;

          // Send app acknowledgement. If application can't hear us
          // we better release the packet 'cause no one else will.
          if (!WinDdePostMsg(pObjIDA->hwndSender, hwnd,
                             WM_DDE_ACK, pDdeStruct, TRUE)) {
            DosFreeMem(pDdeStruct);
            if (pObjIDA->bAllErrors)
              WinPostMsg(pObjIDA->hwndClient, UM_LINKERROR,
                         MPFROMP(pObjIDA),
                         MPFROM2SHORT(APPNOTRESPONDING, WM_DDE_ACK));
          }
        }
      }
      break; // end of WM_DDE_UNADVISE

    //
    // Process a timer tick.
    case WM_TIMER:

      // This is the main part of WM_DDE_ADVISE processing.

      // Get the windows server's object window and server's
      // client window IDAs.
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

      // If we aren't required to wait for the sender to
      // acknowledge receipt of a DDESTRUCT data packet or the
      // sender has already acknowledged receipt of the last
      // DDESTRUCT data packet...
      if (!pObjIDA->bAcknowledgeRequest ||
          !pObjIDA->bAcknowledgementPending)

        // ...and FNODATA isn't enabled, pack up time else...
        if (!pObjIDA->bNotifyUpdate) {
          PCHAR szTime = getTime();
          pDdeStruct = createDataPacket(pObjIDA->hwndSender,
                                        (PSZ)pObjIDA->pszItemName,
                                        0L, (PSZ)szTime,
                                        USHORT(strlen(szTime)+1));
        }
        else

          // ...if the DDE_FNODATA bit is set, we ship a data packet
          // with zero length data to notify the sender that updated
          // data is available. The sender is then expected to
          // actually get the data via a WM_DDE_REQUEST.
          if (pObjIDA->bNotifyUpdate) {
            pDdeStruct = createDataPacket(pObjIDA->hwndSender,
                                          (PSZ)pObjIDA->pszItemName);
            pDdeStruct->offabData = pDdeStruct->offszItemName = 0;
            pDdeStruct->fsStatus |= DDE_FNODATA;
          }


      // If the DDE_FACKREQ bit of fsStatus was set in the
      // DDESTRUCT data packet received by the WM_DDE_ADVISE
      // message, the sender wants us to set the
      // "acknowledgement-requested" bit in the returned
      // DDESTRUCT packet.
      if (pObjIDA->bAcknowledgeRequest)
        pDdeStruct->fsStatus |= DDE_FACKREQ;

      // WM_TIMER processing is always a response to a
      // WM_DDE_ADVISE message.
      pDdeStruct->fsStatus &= ~DDE_FRESPONSE;

      // At this point the packet is complete and:
      //  DDESTRUCT->fsStatus:
      //     DDE_FACREQ =    If set, client receiver should return
      //                     an WM_DDE_ACK message
      //     DDE_FRESPONSE = If set, data is a reponse to a
      //                     WM_DDE_REQUEST else it is a response
      //                     to a WM_DDE_ADVISE

      // If application isn't listening, the data packet won't
      // get deleted so...
      if (!WinDdePostMsg(pObjIDA->hwndSender, hwnd, WM_DDE_DATA,
                         pDdeStruct, TRUE)) {

        // ...we do it here.
        DosFreeMem(pDdeStruct);
        if (pObjIDA->bAllErrors)
          WinPostMsg(pObjIDA->hwndClient, UM_LINKERROR,
                     MPFROMP(pObjIDA),
                     MPFROM2SHORT(APPNOTRESPONDING, WM_DDE_ACK));
      }
      break; // end of WM_TIMER

    //
    // Unsolicited data received.
    //   mp1 = sender's window handle
    //   mp2 = pointer to DDESTRUCT instance
    // Receiver is expected to reply with a positive WM_DDE_ACK
    // message (DDE_FACK bit = 1) if it accepts the unsolicited
    // data, or with a negative WM_DDE_ACK (DDE_FACK bit = 0)
    // if it does not.
    case WM_DDE_POKE:
      {
        //  Alarm clock uses a poke to change the time update
        // interval.

        // Get sender's window handle, the DDE data packet and
        // and the server's IDA.
        hwndSender = HWNDFROMMP(mp1);
        pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);
        pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

        // Make the elitist assumtions that we will reject
        // the request but that its not because we are too busy.
        pDdeStruct->fsStatus = USHORT((pDdeStruct->fsStatus&DDE_FACK)|
                                                     ~DDE_FBUSY);

        //  If "I am for you"...
        if (pObjIDA->hwndSender == hwndSender) {

          // ...and this is a request for a display interval
          // change (which we are able to handle)...
          if (!strcmpi((PCHAR)DDES_PSZITEMNAME(pDdeStruct),
                       "INTERVAL")) {

            // ...convert data to an interger in the object
            // window instance data area.
            PCHAR temp = (PCHAR)DDES_PABDATA(pDdeStruct);
            INT interval = atoi(temp);

            // We do not support any interval less than a second!
            pObjIDA->interval = /*interval<1000 ? 1000:*/interval;

            // If the timer is running (TRUE from WinStopTimer)...
            habServer = WinQueryAnchorBlock(hwnd);
            if (WinStopTimer(habServer, hwnd, pObjIDA->usTimerNo))

              // ...restart it at the new interval.
              WinStartTimer(habServer, hwnd, pObjIDA->usTimerNo,
                            pObjIDA->interval);

            // Set positive ack flag (otherwise our elitist
            // assumptions will have to stand).
            pDdeStruct->fsStatus |= DDE_FACK;

            // Report action to client window.
            WinPostMsg(pObjIDA->hwndClient, UM_LINKMESSAGE,
                       MPFROMP(pObjIDA), MPFROMSHORT(WM_DDE_POKE));
          }
        }

        // Acknowledge we got the message; if DDE_FACK is 1 and
        // DDE_FBUSY is 0, not only did we get the message but
        // we actually accomplished the interval change. The client
        // is responsible for dumping the data packet back into
        // the memory pool.
        if (!WinDdePostMsg(hwndSender, hwnd, WM_DDE_ACK,
                           pDdeStruct, TRUE)) {
          DosFreeMem(pDdeStruct);
          if (pObjIDA->bAllErrors)
            WinPostMsg(pObjIDA->hwndClient, UM_LINKERROR,
                       MPFROMP(pObjIDA),
                       MPFROM2SHORT(APPNOTRESPONDING, WM_DDE_ACK));
        }
      }
      break;  // WM_DDE_POKE

    //
    //  Execution of command(s) requested.
    //  mp1 = sender's window handle
    //  mp2 = DDESTRUCT pointer
    //    offabData = command(s) to be executed
    case WM_DDE_EXECUTE:
      {
        // The server recognizes 4 commands of the form
        // "ALARMn,xxxxxx" where 'n' is a number from 1 to 4
        // and 'xxxxxx' is a number from 000000 thru 999999.
        // Each "ALARMn" has a distinctive sound and the number
        // 'xxxxxx' specifies how long the alarm will sound.
        // No spaces are allowed anywhere within the command
        // string. 'xxxxxx' doesn't have to be zero padded.

        // We should recognize this by now!
        hwndSender = HWNDFROMMP(mp1);
        pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
        pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);

        // Until proven otherwise, assume we reject request but
        // not because we are too busy.
        pDdeStruct->fsStatus = USHORT((pDdeStruct->fsStatus&DDE_FACK)|
                                                     ~DDE_FBUSY);
        // If sender is our sender attempt the alarm.
        if (pObjIDA->hwndSender == hwndSender) {

          // The only WM_DDE_EXECUTE item we understand is an
          // "ALARM"; if the "ALARM" appears in the item name string
          // it is either the alarm item name or the alarm command
          // (some application's ignore the manual and send the
          // command in the item name).
          PCHAR pCmdItem = (PCHAR)DDES_PSZITEMNAME(pDdeStruct);
          if (strstr(pCmdItem, "ALARM")) {

            // If there's data...
            PCMDSTRUCT pCmdStruct;
            if (pDdeStruct->offabData)

              // ...get the command from the data area else...
              pCmdStruct = (PCMDSTRUCT)DDES_PABDATA(pDdeStruct);
            else

              // ...assume the command is in the item name area and...
              pCmdStruct = (PCMDSTRUCT)DDES_PSZITEMNAME(pDdeStruct);

            // ...split it into command and duration.
            *pCmdStruct->separator = 0;

            // Get the alarm number by comparing command against
            // alarm string array; breaks on a match with alarm
            // number in 'i'.
            USHORT i;
            for (i=0; i<commands; i++)
              if (!strcmpi((PCHAR)szCmd[i], (PCHAR)pCmdStruct->szCmdStr))
                break;

            // 'i' should contain a valid (0-3) alarm number; if so...
            if (i<commands) {

              // ...load alarm variables and attempt to spin off
              // an alarm thread for it. On success...
              pObjIDA->usAlarm = i;
              pObjIDA->usDuration = (USHORT)atoi(pCmdStruct->szDuration);
              if (_beginthread(alarmThread, 4096, (PVOID)pObjIDA))

                // ...set the positive acknowledge bit in the data
                // packet's fsStatus word and...
                pDdeStruct->fsStatus |= DDE_FACK;

                // ...report action to client window.
                WinPostMsg(pObjIDA->hwndClient, UM_LINKMESSAGE,
                           MPFROMP(pObjIDA),
                           MPFROMSHORT(WM_DDE_EXECUTE));
            }
          }

          // At this point we must acknowledge our sender's
          // request. It will be a positive acknowledgement if;
          //  - the app requested an alarm
          //  - the alarm was a valid alarm (i.e., 0 thru 4)
          //  - we successfully spun off the alarm thread.
          // Failure in any of these will result in a negative
          // acknowledgement. In either case, the sender is
          // responsible for releasing the data packet back into
          // the memory pool.
          if (!WinDdePostMsg(hwndSender, hwnd, WM_DDE_ACK,
                             pDdeStruct, TRUE)) {
            DosFreeMem(pDdeStruct);
            if (pObjIDA->bAllErrors)
              WinPostMsg(pObjIDA->hwndClient, UM_LINKERROR,
                         MPFROMP(pObjIDA),
                         MPFROM2SHORT(APPNOTRESPONDING, WM_DDE_ACK));
          }
        }
      }
      break;  // WM_DDE_EXECUTE

    //
    // Receipt of data acknowledged.
    //   mp1 = sender's window handle
    //   mp2 = pointer to DDESTRUCT data packet
    //     DDESTRUCT usStatus:
    //       DDE_FACK:         1 = accepted, 0 = rejected
    //       DDE_FBUSY:        0 = not busy, 1 = busy
    //       DDE_FAPPSTATUS:   1 = not understood, ignored
    //
    case WM_DDE_ACK:

      // If the sender has requested receipt acknowledgement, it
      // will send a WM_DDE_ACK message when it is ready for the
      // the next data update. In this case the server shouldn't
      // ship anything until the sender has issued the WM_DDE_ACK.

      // Get sender's window handle, server object window IDA and
      // sender's data packet.
      hwndSender = HWNDFROMMP(mp1);
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);

      if (pObjIDA->hwndSender == hwndSender) {

        // If there is a acknowledge receipt protocol going on...
        if (pObjIDA->bAcknowledgeRequest) {

          // ...ignore DDE_FBUSY; if sender is busy it loses the
          // last data update. However, if the sender rejects our
          // last data update, we sound an alarm but otherwise
          // ignore the condition.
          if (!pDdeStruct->fsStatus&DDE_FACK)
            for (USHORT i=0; i<4; i++)

          // Let everybody know the sender says its ready for
          // another update.
          pObjIDA->bAcknowledgementPending = FALSE;
        }
      }

      // The arrival of a WM_DDE_ACK is always doomsville for the
      // data packet at either the server or application's end of
      // the link. If acknowledge-request is disabled, the
      // WM_DDE_ACK handler will never be invoked and the data
      // packet is freed at the receiver handler for each message
      // type otherwise it is deleted here.
      DosFreeMem(pDdeStruct);
      break;  // end of WM_DDE_ACK

    //
    // Termination of DDE server link requested.
    //   mp1 = sender's window handle
    //   mp2 = reserved
    case WM_DDE_TERMINATE:
      hwndSender = HWNDFROMMP(mp1);
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);

      if (pObjIDA->hwndSender == hwndSender) {

        // Send a Termination notice to the sender. If app isn't
        // listening free data packet.
        if (!WinDdePostMsg(hwndSender, hwnd, WM_DDE_TERMINATE,
                           pDdeStruct, TRUE))
          DosFreeMem(pDdeStruct);

        // ...let the client window know.
        // NOTE: We must 'send' this message; 'post'ing it results
        //       in a crash because 'pObjIDA' has been released
        //       before the client window gets to it.
        WinSendMsg(pObjIDA->hwndClient, UM_LINKMESSAGE,
                   MPFROMP(pObjIDA), MPFROMSHORT(WM_DDE_TERMINATE));

        // Irrespective of whether the app stayed around for the
        // funeral, dump the DDESTRUCT allocation, object window
        // IDA and the object window itself.
        delete pObjIDA->pszItemName;
        delete pObjIDA;
        WinDestroyWindow(hwnd);
      }
      break;  // WM_DDE_TERMINATE

    //
    // Data received.
    //  mp1 = window handle of sender
    //  mp2 = DDESTRUCT data packet pointer
    //  DDESTRUCT->fsStatus:
    //     DDE_FACREQ =    If set, client receiver should return
    //                     an WM_DDE_ACK message
    //     DDE_FRESPONSE = If set, data is a reponse to a
    //                     WM_DDE_REQUEST else it is a response
    //                     to a WM_DDE_ADVISE
    case WM_DDE_DATA:

      // Since we don't know noth'in about WM_DDE_DATA messages on
      // this side of the link, we'll just let mama know we got
      // one and do the housekeeping associated with its receipt.
      hwndSender = HWNDFROMMP(mp1);
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);

      if (pObjIDA->hwndSender == hwndSender) {
        WinPostMsg(pObjIDA->hwndClient, UM_LINKMESSAGE,
                   MPFROMP(pObjIDA), MPFROMSHORT(WM_DDE_DATA));
        if (pDdeStruct->fsStatus&DDE_FACKREQ) {
          if (!WinDdePostMsg(hwndSender, hwnd, WM_DDE_ACK, pDdeStruct,
                             pObjIDA->pflags)) {
            DosFreeMem(pDdeStruct);
          }
        }
        else
          DosFreeMem(pDdeStruct);
      }
      return FALSE;  // WM_DDE_DATA

    //
    // Out there, somewhere, is a server which claims it is for us.
    //   mp1 = window handle of sender
    //   mp2 = DDEINIT data packet pointer
    //    typedef struct _DDEINIT {
    //      ULONG cb;             // sizeof(DDEINIT)
    //      PSZ   pszAppName;     // name of the responding applica-
    //                            // tion.
    //                            // It must not be a zero-length
    //                            // string.
    //      PSZ   pszTopic;       // name of the topic that the server
    //                            // is willing to support; it must
    //                            // not be a zero-length string
    //      ULONG offConvContext; // conversation context
    //    } DDEINIT;
    //  The DDEINIT structure must be in a sharable segment; it is
    //  the responsibility of the receiving window procedure to free
    //  this segment. A mosal window, such as a message box, must not
    //  be posted during the processing of this message.
    case WM_DDE_INITIATEACK:
      {
#pragma warn -aus
        // Since we don't know noth'in about WM_DDE_INITIATEACK
        // messages on this side of the link, we'll just let mama
        // know we got one and do the housekeeping associated with
        // its receipt.
        hwndSender = HWNDFROMMP(mp1);
        pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
        PDDEINIT pDdeInit = (PDDEINIT)PVOIDFROMMP(mp2);

        // It is a federal offense to mess with other peoples mail.
        if (pObjIDA->hwndSender == hwndSender) {
          WinPostMsg(pObjIDA->hwndClient, UM_LINKMESSAGE,
                     MPFROMP(pObjIDA), MPFROMSHORT(WM_DDE_INITIATEACK));
          DosFreeMem(pDdeStruct);
        }
      }
      return FALSE;  // WM_DDE_INITIATEACK

    //
    default:
      mr = WinDefWindowProc(hwnd, msg, mp1, mp2);
      break;
  }

  return mr;
} // end of objWndProc()


/*-------------------------------------------------------------------
   createDataPacket - allocates shared memory for a DDESTRUCT, its
                      item name and data, initializes the elements of
                      the allocation and arranges for that allocation
                      to be accesible to the both the DDE server and
                      sender processes.

   PARAMETERS:  hwndServer -  DDE server window handle
                pszItemName - item name
                usStatus    - DDE exchange status (default = 0)
                pszData    -  data (default = NULL)
                usDataLen  -  length of data (default = 0)
                usFormat    - data format (default = DDEFMT_TEXT)

   RETURNS: pointer to an initialized DDESTRUCT instance.
*/

PDDESTRUCT EXPENTRY createDataPacket(HWND hwndServer,
                                     PSZ pszItemName,
                                     USHORT fsStatus,
                                     PSZ pszData,
                                     USHORT usDataLen,
                                     USHORT usFormat) {

  // Calculate length of item name and, using this, the length
  // of the data passed to us and the size of a bare bones
  // DDESTRUCT instance, calculate the size of the complete
  // DDESTRUCT (i.e., with item name and item data) instance.
  INT usItemLen = USHORT(strlen((PCHAR)pszItemName)+1);
  INT usTotalLen = USHORT(sizeof(DDESTRUCT)+usItemLen+usDataLen);

  // Now allocate a DDESTRUCT of the appropriate dimensions
  // which can be shared (OBJ_GIVEABLE) and...
  PDDESTRUCT pDdeStruct;
  DosAllocSharedMem((PPVOID)&pDdeStruct, NULL, usTotalLen,
                    OBJ_GIVEABLE | PAG_COMMIT | PAG_WRITE);

  // ...load it with the appropriate data.
  pDdeStruct->cbData = usTotalLen;
  pDdeStruct->fsStatus = fsStatus;
  pDdeStruct->usFormat = usFormat;

  // The item name starts immediately after the last byte of
  // the DDESTRUCT instance proper and the data starts immediately
  // after the last byte of the item name.
  pDdeStruct->offszItemName = sizeof(DDESTRUCT);
  pDdeStruct->offabData = USHORT(sizeof(DDESTRUCT) + usItemLen);

  // Since neither the item name nor the data are ASCIIZ, we
  // use memcpy() to put them in place.
  memcpy(DDES_PSZITEMNAME(pDdeStruct), pszItemName, usItemLen);
  memcpy(DDES_PABDATA(pDdeStruct), pszData, usDataLen);

  // Finally, we give the server process access to the DDESTRUCT
  // instance...
  PID pidServer;
  TID tidServer;
  WinQueryWindowProcess(hwndServer, &pidServer, &tidServer);
  DosGiveSharedMem(pDdeStruct, pidServer, PAG_WRITE);

  // ...and return the DDESTRUCT allocation pointer.
  return pDdeStruct;
} // end of createDataPacket)


//-------------------------------------------------------------------
PCHAR EXPENTRY getTime(VOID) {
  struct _DATETIME dt;
  static CHAR szTimeBuf[30];

  DosGetDateTime((LPDT)&dt);
  wDateTime[HOUR]   = (USHORT)dt.hours;
  wDateTime[MINUTE] = (USHORT)dt.minutes;
  wDateTime[SECOND] = (USHORT)dt.seconds;
  sprintf(szTimeBuf, "%02d:%02d:%02d",
          wDateTime[HOUR], wDateTime[MINUTE], wDateTime[SECOND]);

  return szTimeBuf;
} // end of getTime()


//-------------------------------------------------------------------
PCHAR getDate(VOID) {
  struct _DATETIME dt;
  static CHAR szDate[9];

  DosGetDateTime((LPDT)&dt);
  wDateTime[MONTH] = (USHORT)dt.month;
  wDateTime[DAY]   = (USHORT)dt.day;
  wDateTime[YEAR]  = dt.year;           // Year is two bytes
  sprintf(szDate, "%02d-%02d-%02d", dt.month, dt.day, dt.year);

  return szDate;
} // end of getDate()


//------------------------------------------------------------------
VOID alarmThread(PVOID x) {
  POBJIDA pObjIDA = (POBJIDA)x;
  USHORT j;

  switch(pObjIDA->usAlarm) {
    case 0:
      for (j=0; j<pObjIDA->usDuration; j++) {
        DosBeep(1300, 10);
        DosBeep(800, 10);
        DosBeep(2000, 50);
      }
      break;
    case 1:
      for (j=0; j<pObjIDA->usDuration; j++) {
        DosBeep(800, 10);
        DosBeep(900, 10);
        DosBeep(1000, 10);
        DosBeep(1100, 10);
        DosBeep(1200, 10);
        DosBeep(1300, 10);
        DosBeep(1400, 10);
        DosBeep(1500, 10);
        DosBeep(1600, 10);
        DosBeep(1600, 10);
        DosSleep(0);
      }
      break;
    case 2:
      for (j=0; j<pObjIDA->usDuration; j++) {
        DosBeep(4000, 20);
        DosBeep(1000, 30);
        DosBeep(3000, 40);
        DosBeep(500, 10);
        DosSleep(0);
      }
      break;
    case 3:
      for (j=0; j<pObjIDA->usDuration; j++) {
        DosBeep(6000, 100);
        DosSleep(0);
      }
      break;
  }

  _endthread();
} // end of alarmThread()

//-------------------------------------------------------------------
BOOL setListboxText(POBJIDA pObjIDA) {
  BOOL ret = FALSE;

  SHORT count = (SHORT)WinSendMsg(pObjIDA->hwndListbox,
                                  LM_QUERYITEMCOUNT, 0L, 0L);

  POBJIDA temp;
  CHAR szText[80];
  for (USHORT i=0; i<count; i++) {
    temp = (POBJIDA)WinSendMsg(pObjIDA->hwndListbox,
                               LM_QUERYITEMHANDLE,
                               MPFROMSHORT(i), 0L);

    if (temp->hwndSender == pObjIDA->hwndSender) {
      sprintf(szText, " (%p) %s", pObjIDA->hwndSender,
                                  pObjIDA->szText);
      ret = (BOOL)WinSendMsg(pObjIDA->hwndListbox, LM_SETITEMTEXT,
                             MPFROMSHORT(i), szText);
    }
  }

  return ret;
} // end of setListboxText()

/*-------------------------------------------------------------------
   toggleCheck - checks menu option 'itemID' if it is unchecked,
                 unchecks it if it is checked.

   PARAMETERS:
     HWND  hwndMenu - topmost menu handle
     ULONG itemID   - menu item id

   RETURNS:  TRUE if item is checked, FALSE if item is not checked
             after toggle operation is performed.
*/

BOOL toggleCheck(HWND hwndMenu, ULONG itemID) {

  // Get check state of menu option,...
  USHORT bChecked = (USHORT)WinSendMsg(hwndMenu, MM_QUERYITEMATTR,
                                       MPFROM2SHORT(itemID, TRUE),
                                       MPFROMSHORT(MIA_CHECKED));

  bChecked = USHORT(bChecked?1:0);

  // ...try to toggle it. If attempt failed...
  if (!WinSendMsg(hwndMenu, MM_SETITEMATTR,
                  MPFROM2SHORT(itemID, TRUE),
                  MPFROM2SHORT(MIA_CHECKED, bChecked?~MIA_CHECKED:
                                                      MIA_CHECKED))){
    // ...do an alarm and report same state else...
    return bChecked;
  }

  // ...return toggle of previous check state.
  return !bChecked;
} // end of toggleCheck()

