/* TICKOSC.CPP: DDE Alarm clock testbed client application

   This is the client/application side of the DDE demo. Unlike a "real"
   application, TICKOSC provides options that permit the operator to
   experiment with the standard DDE communications prototcols. The
   following protocol options appear on the "Setup" menu:

   FACKREQ - If checked, server will not send new data until client
             application acknowledges receipt of last data.
   FNODATA - If checked, server sends a zero-length data packet to
             signal that new data is available and the client
             application gets the data (when its good and ready) via
             a WM_DDE_REQUEST.
   RETRY   - If checked, application will try to send a DDE message
             until it is successful; if not checked, any failure to
             send a DDE message will be reported as an error.

   This demo uses all of the DDE messages:

    WM_DDE_INITIATE and WM_DDE_INITIATE_ACK are used to establish
    the DDE link.

    WM_DDE_DATA to get data.

    WM_DDE_REQUEST, WM_DDE_ADVISE and WM_DDE_UNADVISE are used to
    request data.

    WM_DDE_POKE is used to change the display frequency.

    WM_DDE_EXECUTE is used to request an alarm and specify its
    duration.

    WM_DDE_TERMINATE is used to terminate the DDE link.

   Besides being an example of DDE usage, this is an example of
   bidirectional communication between windows; locally, between
   the client and server window and remotely, via a DDE link, between
   the client application and the Alarm Clock Server.

   The client window handles all interaction between the operator
   and the object window handles all interaction between the client
   application and the server. The client and object windows commun-
   icate with each other to control the DDE link exchange.

 ====================================================================
   NOTES:  (1) TICKOSC requires the following files:
                 TICKOSC.CPP  - main(), window proc & functions
                 TICKOSC.H    - constants, structures & prototypes
                 TICKOSC.RC   - resources
           (2) In the comments I refer to a data packet; this is
               always a reference to the DDESTRUCT instance which is
               used as a vehicle for passing messages between the
               application/server and server/application.
           (3) The extensive commenting may be considered a bit
               too much by some; however, I intended this source
               to be a working compendium of all the information
               that is spread over at least two manuals in the
               IBM documentation and to provide an alternative
               phrasing of the text of these manuals.
           (4) The IBM documentation (both online and hardcopy)
               consistently refers to the DDESTRUCT data member
                "usStatus". This is an error; the member is actually
                "fsStatus".

  -------------------------------------------------------------------
     PM FUNCTIONS DEMONSTRATED:
       WinInitialize()              WinCreateMsgQueue()
       WinRegisterClass()           WinCreateStdWindow()
       WinMessageBox()              WinDestroyMsgQueue()
       WinTerminate()               WinQueryWindowRect()
       WinSetWindowPos()            WinGetMsg()
       WinDispatchMsg()             WinDestroyWindow()
       WinQueryWindow()             WinWindowFromID()
       WinSendMsg()                 WinDlgBox()
       WinSetWindowPtr()            WinQueryWindowPtr()
       WinBeginPaint()              WinDlgBox()
       WinDrawText()                WinEndPaint()
       WinInvalidateRect()          WinDdeInitiate()
       WinPostMsg()                 WinDefWindowProc()
       WinDdeRespond()              WinDdePostMsg()
       DosFreeMem()                 WinStartTimer()
       WinStopTimer()               WinQueryAnchorBlock()
       DosAllocSharedMem()          WinQueryWindowProcess()
       DosGiveSharedMem()           DosGetDateTime()
       DosBeep()                    DosSleep()
       WinSetWindowText()           WinSetDlgItemText()
       WinDismissDlg()              WinQueryDlgItemText()

     PM MESSAGES DEMONSTRATED:
       MM_SETITEMATTR               WM_PAINT
       WM_CREATE                    WM_COMMAND
       WM_CLOSE                     WM_DDE_INITIATEACK
       WM_DDE_REQUEST               WM_DDE_ADVISE
       WM_DDE_UNADVISE              WM_DDE_POKE
       WM_DDE_EXECUTE               WM_DDE_TERMINATE
       WM_DDE_ACK                   WM_WUERYITEMATTR

  -------------------------------------------------------------------
               Copyright (c) 1993 Borland International
  -------------------------------------------------------------------
     04-04-93 EAS   : i hate initial releases
*/

#include <string.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>

#define INCL_WIN
#define INCL_DOS

#include <os2.h>

#include "tickosc.h"

PCHAR szItemName[] =  { "TIME", "INTERVAL", "ALARM" };

PCHAR szLink[] = { "NO LINK", "INITIALIZING...", "LINKED" };
PCHAR adreq[] =  { " REQUEST", " ADVISE" };
PCHAR sdel[] =   { ": SERVER DELETES", ": CLIENT DELETES" };
PCHAR rf[] =     { "", " RETRY " };
PCHAR freq[] =   { "", " FACKREQ " };
PCHAR nodata[] = { "", " FNODATA " };

HWND hwndClient;

//*******************************************************************
INT main(VOID) {
  HAB   hab;
  HMQ   hmq;
  ULONG flCreateFlags = FCF_TITLEBAR      | FCF_SYSMENU  |
                        FCF_SIZEBORDER    | FCF_MINMAX   |
                        FCF_SHELLPOSITION | FCF_TASKLIST |
                        FCF_MENU          | FCF_ICON;
  QMSG  qmsg;
//HWND  hwndClient;

  hab = WinInitialize(0);
  hmq = WinCreateMsgQueue(hab, 0);
  WinRegisterClass(hab,(PSZ)szClientClass, clientWndProc,
                   CS_SIZEREDRAW, sizeof(PCLIENTIDA));

  WinRegisterClass(hab,(PSZ)szObjClass, objWndProc,
                   0L, sizeof(POBJIDA));

  HWND hwndFrame = WinCreateStdWindow(HWND_DESKTOP, 0L,
                                      &flCreateFlags,
                                      (PSZ)szClientClass, NULL,
                                      0L, 0L, rid, &hwndClient);

  WinSetWindowPos(hwndFrame, HWND_TOP, 5, 5, 205, 65,
                  SWP_SIZE|SWP_MOVE|SWP_SHOW|SWP_ACTIVATE);

  while(WinGetMsg(hab, &qmsg, 0L, 0, 0))
    WinDispatchMsg(hab, &qmsg);

  WinDestroyWindow(hwndFrame);
  WinDestroyMsgQueue(hmq);
  WinTerminate(hab);

  return 0;
} // end of main()


/*------------------------------------------------------------------
   clientWndProc - handles user input and initializes the DDE
                   exchange.
*/

MRESULT EXPENTRY clientWndProc(HWND hwnd, ULONG msg,
                               MPARAM mp1, MPARAM mp2) {
  MRESULT     mr = FALSE;      // PM function call result
  BOOL        bb;              // utility logic
  PCLIENTIDA  pClientIDA;      // client window instance data area

  switch(msg) {

    //
    case WM_CREATE:
      {
        // Allocate client window's instance data area, initialize
        // strings and menu handle.
        pClientIDA = new CLIENTIDA;
        strcpy(pClientIDA->szTime, "00:00:00");
        strcpy(pClientIDA->szInterval, "1000");
        pClientIDA->hwndMenu = WinWindowFromID(WinQueryWindow(hwnd,
                                             QW_PARENT), FID_MENU);

        // Initialize checked "Setup" menu options.
        WinSendMsg(pClientIDA->hwndMenu, MM_SETITEMATTR,
                   MPFROM2SHORT(IDM_NOFREE, TRUE),
                   MPFROM2SHORT(MIA_CHECKED, ~MIA_CHECKED));
        strcpy(pClientIDA->szNOFREE, sdel[0]);

        WinSendMsg(pClientIDA->hwndMenu, MM_SETITEMATTR,
                   MPFROM2SHORT(IDM_RETRY, TRUE),
                   MPFROM2SHORT(MIA_CHECKED, MIA_CHECKED));
        strcpy(pClientIDA->szRETRY, rf[1]);
        pClientIDA->usDDEflags |= DDEPM_RETRY;

        WinSendMsg(pClientIDA->hwndMenu, MM_SETITEMATTR,
                   MPFROM2SHORT(IDM_FACKREQ, TRUE),
                   MPFROM2SHORT(MIA_CHECKED, MIA_CHECKED));
        strcpy(pClientIDA->szFACKREQ, freq[1]);
        pClientIDA->usDDEflags |= DDE_FACKREQ;

        WinSendMsg(pClientIDA->hwndMenu, MM_SETITEMATTR,
                   MPFROM2SHORT(IDM_FNODATA, TRUE),
                   MPFROM2SHORT(MIA_CHECKED, ~MIA_CHECKED));
        strcpy(pClientIDA->szFNODATA, nodata[0]);

        // Default display interval is 1 second.
        strcpy(pClientIDA->szInterval, "1000");
        WinSetWindowPtr(hwnd, QWL_USER, pClientIDA);

        // Attempt to establish the default server link.
        initObjWindow(hwnd);
        setTitle(hwnd);
      }
      break;

    //
    case WM_PAINT:
      {
        RECTL rcl;

        WinQueryWindowRect(hwnd, &rcl);
        HPS hps = WinBeginPaint(hwnd, 0L, 0L);

        // Extract current client window IDA and ...
        pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);

        // ...display time.
        WinDrawText(hps, -1, (PSZ)pClientIDA->szTime,
                    &rcl, clrForeground, clrBackground,
                    DT_CENTER|DT_VCENTER|DT_ERASERECT);

        WinEndPaint(hps);
      }
      break;

    //
    // Object window says there's no server out there that can
    // do the job so...
    case UM_LINKINITFAILURE:

      // ...sound alarm and display error condition message.
      setDDE_Controls(hwnd);
      WinMessageBox(HWND_DESKTOP, hwnd,
                    (PSZ)"Can't find Alarm Clock server",
                    (PSZ)"", 0L, MB_OK|MB_MOVEABLE);

      // Update titlebar to reflect linkless condition.
      pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      strcpy(pClientIDA->szLink, szLink[0]);
      setTitle(hwnd);
      break;

    //
    // Object window says that server is not responding so...
    case UM_NOSERVERRESPONSE:
      {
        switch(SHORT1FROMMP(mp1)) {

          case UM_DDE_TIMEOUT:
            WinMessageBox(HWND_DESKTOP, hwnd,
                          (PSZ)"Server transmission timeout",
                          (PSZ)"", 0L, MB_OK|MB_MOVEABLE);
            break;

          case WM_DDE_ACK:
            WinMessageBox(HWND_DESKTOP, hwnd,
                          (PSZ)"Could not send ACK to server",
                          (PSZ)"", 0L, MB_OK|MB_MOVEABLE);
            break;

          default:
            WinMessageBox(HWND_DESKTOP, hwnd,
                          (PSZ)"Server not responding",
                          (PSZ)"", 0L, MB_OK|MB_MOVEABLE);
            break;
        }

        // Update titlebar to reflect linkless condition.
        setDDE_Controls(hwnd);
        pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
        strcpy(pClientIDA->szLink, szLink[0]);
        setTitle(hwnd);
      }
      break;

    //
    // Object window says DDE link established so...
    case UM_LINKED:

      // ...update DDE control options in "Setup" menu and...
      setDDE_Controls(hwnd, TRUE);

      // ...titlebar.
      pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      strcpy(pClientIDA->szLink, szLink[2]);
      setTitle(hwnd);
      break;

    //
    // Object window says DDE link has been terminated so...
    case UM_TERM:

      // ...update DDE control options and...
      setDDE_Controls(hwnd);

      // ...titlebar.
      pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      strcpy(pClientIDA->szLink, szLink[0]);
      setTitle(hwnd);
      break;


    case WM_COMMAND:

      switch(SHORT1FROMMP(mp1)) {

        //
        // Operator requests DDE link; create object window and
        // update titlebar ("Initializing...").
        case IDM_INIT:
          initObjWindow(hwnd);
          setTitle(hwnd);
          break;

        //
        // Operator toggled retry-if-queue-is-full state.
        case IDM_RETRY:
          {
            // Get client window IDA,...
            pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);

            // ...toggle retry state and load appropriate string.
            bb = toggleCheck(pClientIDA->hwndMenu, IDM_RETRY);
            strcpy(pClientIDA->szRETRY, rf[bb]);

            // If an object window is available, tell object window
            // that retry state has changed and pass it retry state
            // in mp1.
            // NOTE: If an object window isn't yet available, both
            //       'pflag' and 'uflag' bits are set based on the
            //       state of the corresponding "Setup" menu option
            //       checks when the object window is created.
            if (pClientIDA->hwndObj)
              WinPostMsg(pClientIDA->hwndObj, UM_RETRY,
                         MPFROMSHORT(bb), 0L);

            setTitle(hwnd);
          }
          break;

        //
        // Operator toggled server-deletes-data-packet state.
        case IDM_NOFREE:
          {
            // Get client window IDA and...
            pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);

            // ...toggle server-deletes state and load appropriate
            // string to client IDA.
            bb = toggleCheck(pClientIDA->hwndMenu, IDM_NOFREE);
            strcpy(pClientIDA->szNOFREE, sdel[bb]);

            // If an object window is available, let it know that
            // the server-deletes state has changed and pass new
            // state in mp1.
            if (pClientIDA->hwndObj)
              WinPostMsg(pClientIDA->hwndObj, UM_NOFREE,
                         MPFROMSHORT(bb), 0L);

            setTitle(hwnd);
          }
          break;

        //
        // Operator toggled send-data-when-client-acknowledges-last-
        // data state.
        case IDM_FACKREQ:
          {
            // Get client window IDA and...
            pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);

            // ...toggle DDE_FACKREQ state and load appropriate
            // string to client IDA.
            bb = toggleCheck(pClientIDA->hwndMenu, IDM_FACKREQ);
            strcpy(pClientIDA->szFACKREQ, freq[bb]);

            // Tell object window that DDE_FACKREQ state has
            // changed and pass new state in mp1.
            if (pClientIDA->hwndObj)
              WinPostMsg(pClientIDA->hwndObj, UM_FACKREQ,
                         MPFROMSHORT(bb), 0L);

            setTitle(hwnd);
          }
          break;

        //
        // Operator toggled server-sends-0 length-data-to-signal-
        // new-data-available state.
        case IDM_FNODATA:
          {
            // Get client window IDA and...
            pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);

            // ...toggle DDE_FNODATA state and load appropriate
            // string to client IDA.
            bb = toggleCheck(pClientIDA->hwndMenu, IDM_FNODATA);
            strcpy(pClientIDA->szFNODATA, nodata[bb]);

            // Tell object window that DDE_FNODATA state has
            // changed and pass new state in mp1.
            if (pClientIDA->hwndObj)
              WinPostMsg(pClientIDA->hwndObj, UM_FNODATA,
                         MPFROMSHORT(bb), 0L);

            setTitle(hwnd);
          }
          break;

        case IDM_ALARM1:
          pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
          WinPostMsg(pClientIDA->hwndObj, UM_EXECUTE,
                     MPFROM2SHORT(1, 10), 0L);
          break;

        case IDM_ALARM2:
          pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
          WinPostMsg(pClientIDA->hwndObj, UM_EXECUTE,
                     MPFROM2SHORT(2, 20), 0L);
          break;

        case IDM_ALARM3:
          pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
          WinPostMsg(pClientIDA->hwndObj, UM_EXECUTE,
                     MPFROM2SHORT(3, 10), 0L);
          break;

        case IDM_ALARM4:
          pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
          WinPostMsg(pClientIDA->hwndObj, UM_EXECUTE,
                     MPFROM2SHORT(4, 20), 0L);
          break;

        //
        case IDM_ONE:
          pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
          WinPostMsg(pClientIDA->hwndObj, UM_REQUEST, 0L, 0L);
          break;

        //
        // Operator changed update interval so we "poke" the interval
        // change at the server.
        case IDM_POKE:
          WinDlgBox(HWND_DESKTOP, hwnd, (PFNWP)intervalDlgProc, 0L,
                    DLG_INTERVAL, (PVOID)NULL);
          pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);

          if (pClientIDA->hwndObj)
            WinPostMsg(pClientIDA->hwndObj, UM_POKE,
                       MPFROMP((PVOID)pClientIDA->szInterval), 0L);
          break;

        //
        case IDM_START:
          pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
          WinPostMsg(pClientIDA->hwndObj, UM_START, 0L, 0L);
          break;

        //
        case IDM_STOP:
          pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
          WinPostMsg(pClientIDA->hwndObj, UM_STOP, 0L, 0L);
          break;

        //
        case IDM_TERM:
          pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);
          WinPostMsg(pClientIDA->hwndObj, UM_TERM, 0L, 0L);
          pClientIDA->hwndObj = 0;
          break;

        //
        case IDM_EXIT:
          WinPostMsg(hwnd, WM_CLOSE, 0L, 0L);
          break;

        //
        case IDM_ABOUT:
          WinMessageBox(HWND_DESKTOP, hwnd,
                        (PSZ)"DDE Alarm Clock Client",
                        (PSZ)"",
                        1, MB_OK|MB_APPLMODAL);
          break;

        //
        default:
          mr = WinDefWindowProc(hwnd, msg, mp1, mp2);
          break;
      }
      break;

    //
    case WM_CLOSE:
      if (WinMessageBox(HWND_DESKTOP, hwnd,
                       (PSZ)"Exit DDE Alarm Clock?",
                       (PSZ)"",
                       1, MB_YESNO|MB_ICONQUESTION|
                       MB_DEFBUTTON2|MB_APPLMODAL) == MBID_YES) {
        pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd, QWL_USER);

        WinDestroyWindow(pClientIDA->hwndObj);
        WinPostMsg(hwnd, WM_QUIT, 0L, 0L);
      }
      break;

    //
    default:
      mr = WinDefWindowProc(hwnd, msg, mp1, mp2);
      break;
  }
  return mr;
} // end of clientWndProc()


/*-------------------------------------------------------------------
   initObjWindow - creates an object window and initializes object
                   window and client window IDA's to support full
                   birdirectional communuication.
*/

VOID initObjWindow(HWND hwndClient) {
  HWND       hwndObj;       // object window handle
  PCLIENTIDA pClientIDA;    // client window instance data area
  POBJIDA    pObjIDA;       // object window instance data area

  pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwndClient, QWL_USER);
  strcpy(pClientIDA->szLink, szLink[1]);

  // Allocate memory for object window IDA, initialize alarms
  // counter and load client window handle, default application
  // name and default topic.
  pObjIDA = new OBJIDA;
  pObjIDA->hwndClient = hwndClient;
  strcpy(pObjIDA->szappname, szAppName);
  strcpy(pObjIDA->sztopic, szTopic);

  // Transfer initial state of DDE flags from client window IDE
  // to appropriate object window.
  pObjIDA->uflags |= USHORT((pClientIDA->usDDEflags&DDEPM_RETRY));
  pObjIDA->uflags |= USHORT((pClientIDA->usDDEflags&DDEPM_NOFREE));
  pObjIDA->pflags |= USHORT((pClientIDA->usDDEflags&DDE_FACKREQ));
  pObjIDA->pflags |= USHORT((pClientIDA->usDDEflags&DDE_FNODATA));

  // Create object window (WM_CREATE processing of objWndProc()
  // gets IDA pointer in MP1 and loads IDA pointer and object
  // window's handle into its IDA before we return from this
  // call) and...
  hwndObj = WinCreateWindow(HWND_OBJECT, (PSZ)szObjClass, NULL,
                            0L, 0, 0, 0, 0, hwndClient, HWND_TOP,
                            0L, pObjIDA, NULL);

  // ...plug object window's handle into the client window IDA.
  pClientIDA->hwndObj = hwndObj;
} // end of initObjWindow()


/*-------------------------------------------------------------------
   objWndProc - handles the following DDE messages:

    WM_DDE_ACK:         app's response to acknowledgement of a
                        DDE_WM_POKE or DDE_WM_EXECUTE message; if the
                        server rejects our message or is too busy to
                        deal with it we display status in last
                        message status line and delete the data packet
                        else we just delete the data packet.
    WM_DDE_INITIATEACK: completes initialization of DDE link
    WM_DDE_REQUEST:     ignored
    WM_DDE_POKE:        ignored
    WM_DDE_ADVISE:      ignored
    WM_DDE_UNADVISE:    ignored
    WM_DDE_EXECUTE:     ignored
    WM_DDE_TERMINATE:   frees the DDE link object window
    WM_DDE_DATA:        We don't do anything with WM_DDE_DATA execpt
                        notify the client window of its receipt.

    NOTE: This procedure handles all DDE link messages EXCEPT a
          WM_DDE_INITIATE which must be handled by the client window
          procedure of the initiating process.
*/

MRESULT EXPENTRY objWndProc(HWND hwnd, ULONG msg,
                            MPARAM mp1, MPARAM mp2) {
  MRESULT    mr = (MRESULT)FALSE;
  HAB        hab;
  PDDEINIT   pDdeInit;
  PDDESTRUCT pDdeStruct;
  POBJIDA    pObjIDA;
  PCLIENTIDA pClientIDA;

  switch(msg) {

    //
    case WM_CREATE:

      // Get object window's IDA and stuff IDA pointer (passed in
      // 'mp1' by the initObjWindow() WinCreateWindow() call) into
      // the object window's user slot.
      pObjIDA = (POBJIDA)PVOIDFROMMP(mp1);
      WinSetWindowPtr(hwnd, QWL_USER, pObjIDA);

      // Flag a DDE initialization and allow 2 seconds to establish
      // make link.
      pObjIDA->usInitiate = 2000;

      // Now that the formalities are out of the way, request a DDE
      // link from 'szAppName' server for the services provided by
      // 'szTopic'; by default "TICKOS" is the server app name and
      // "CLOCK" is the topic.
      WinDdeInitiate(hwnd, (PSZ)szAppName, (PSZ)szTopic, 0L);

      // Get anchor block and start initialization timer.
      hab = WinQueryAnchorBlock(hwnd);
      WinStartTimer(hab, hwnd, TID_INITIATE, pObjIDA->usInitiate);
      break;

    //
    // Out there, somewhere, is a server which claims it is for us.
    //   mp1 = window handle of sender
    //   mp2 = DDEINIT data packet pointer
    //    typedef struct _DDEINIT {
    //      ULONG cb;             // sizeof(DDEINIT)
    //      PSZ   pszAppName;     // name of the responding applica-
    //                            // tion.
    //                            // It must not be a zero-length
    //                            // string.
    //      PSZ   pszTopic;       // name of the topic that the server
    //                            // is willing to support; it must
    //                            // not be a zero-length string
    //      ULONG offConvContext; // conversation context
    //    } DDEINIT;
    //  The DDEINIT structure must be in a sharable segment; it is
    //  the responsibility of the receiving window procedure to free
    //  this segment. A mosal window, such as a message box, must not
    //  be posted during the processing of this message.
    case WM_DDE_INITIATEACK:

      // Get initialization data packet and object window IDA.
      pDdeInit = (PDDEINIT)PVOIDFROMMP(mp2);
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

      // If we have the correct server and topic...
      if ((!strcmp((PCHAR)pDdeInit->pszAppName, (PCHAR)szAppName)) &
         (!strcmp((PCHAR)pDdeInit->pszTopic, (PCHAR)szTopic))) {

        // ...then we have a link so flag initialization off.
        // Save server and client window handles in the object
        // window IDA and...
        pObjIDA->usInitiate = 0;
        pObjIDA->hwndServer = HWNDFROMMP(mp1);

        // ...let client window know we have the link and turn
        // on the link controls.
        WinPostMsg(pObjIDA->hwndClient, UM_LINKED, 0L, 0L);

        // Since we are the last receiver in the link initialization
        // messaging sequence we delete initialization data packet.
        DosFreeMem(pDdeInit);
      }
      break;

    //
    // Operator has changed retry status; adjust utility data.
    case UM_RETRY:
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      if (mp1)
        pObjIDA->uflags |= DDEPM_RETRY;   // RETRY enabled
      else
        pObjIDA->uflags &= ~DDEPM_RETRY;  // RETRY disabled
      break;

    //
    // Operator has changed server-deletes status; adjust utility
    // data.
    case UM_NOFREE:
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      if (mp1)
        pObjIDA->uflags |= DDEPM_NOFREE;  // NOFREE enabled
      else
        pObjIDA->uflags &= ~DDEPM_NOFREE; // NOFREE disabled
      break;

    //
    // Operator has changed don't-send-data-until-application-
    // acknowledges-last-data-received flag; adjust protocol data.
    case UM_FACKREQ:
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      if (mp1)
        pObjIDA->pflags |= DDE_FACKREQ;   // FACKREQ enabled
      else
        pObjIDA->pflags &= ~DDE_FACKREQ;  // FACKREQ disabled
      break;

    //
    // Operator has changed don't-send-data-until-client
    // acknowledges-last-data-received flag; adjust protocol data.
    case UM_FNODATA:
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      if (mp1)
        pObjIDA->pflags |= DDE_FNODATA;   // NODATA enabled
      else
        pObjIDA->pflags &= ~DDE_FNODATA;  // NODATA disabled
      break;

    //
    // Operator has requested a one shot data update.
    case UM_REQUEST:

      // Get object window IDA, server window handle and flag
      // a one shot data update with a max delivery time of 2
      // seconds.
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      pObjIDA->usRequest = 2000;

      // Allocate a DDE data packet shared between client and
      // server and...
      pDdeStruct = createDataPacket(pObjIDA->hwndServer,
                                    (PSZ)szItemName[TIME]);

      // ...request a data update. If server can't respond...
      if (!WinDdePostMsg(pObjIDA->hwndServer, hwnd, WM_DDE_REQUEST,
                         pDdeStruct, pObjIDA->uflags))

        // ...free data packet allocation else...
        DosFreeMem(pDdeStruct);
      else {

        // ...start data-request timer. If a WM_DDE_DATA isn't
        // received within pObjIDA->usRequest milliseconds, we'll
        // get a "Server not responding" message otherwise we'll
        // get the requested data.
        hab = WinQueryAnchorBlock(hwnd);
        WinStartTimer(hab, hwnd, TID_REQUEST, pObjIDA->usRequest);
      }
      break;

    //
    // Client window requests start of continuous update.
    case UM_START:

      // Get object window IDA, server and client window handles.
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

      // Disable 'Start' option, enable 'Stop' option and flag
      // 'Advise' processing in effect.
      enableMenuItem(pObjIDA->hwndClient, IDM_STOP, TRUE);
      enableMenuItem(pObjIDA->hwndClient, IDM_START, FALSE);

      // Give us a data packet for TIME and...
      pDdeStruct = createDataPacket(pObjIDA->hwndServer,
                                    (PSZ)szItemName[TIME],
                                    pObjIDA->pflags);

      // ...ship it. If server can't hear us, data packet won't get
      // deleted unless we do it here.
      if (!WinDdePostMsg(pObjIDA->hwndServer, hwnd, WM_DDE_ADVISE,
                         pDdeStruct, pObjIDA->uflags))
        DosFreeMem(pDdeStruct);
      break;

    //
    // Client window requests end to continuous data update
    case UM_STOP:

      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

      enableMenuItem(pObjIDA->hwndClient, IDM_START, TRUE);
      enableMenuItem(pObjIDA->hwndClient, IDM_STOP, FALSE);

      pDdeStruct = createDataPacket(pObjIDA->hwndServer,
                                    (PSZ)szItemName[TIME],
                                    pObjIDA->pflags);

      // If server can't hear us packet won't get deleted unless
      // we do it here.
      if (!WinDdePostMsg(pObjIDA->hwndServer, hwnd, WM_DDE_UNADVISE,
                         pDdeStruct, pObjIDA->uflags))
        DosFreeMem(pDdeStruct);
      break;

    //
    // Send an unsolicited message to the server.
    case UM_POKE:
      {
        // The alarm clock server expects an "INTERVAL" item and
        // an ASCII number equal to or greater than 1000. The
        // number is used to set the interval between data updates.

        // Create an interval DDE data packet and poke it at
        // the server.
        PCHAR szData = (PCHAR)PVOIDFROMMP(mp1);
        pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

        pDdeStruct = createDataPacket(pObjIDA->hwndServer,
                                      (PSZ)szItemName[INTERVAL],
                                      pObjIDA->pflags, (PSZ)szData,
                                      USHORT(strlen(szData)+1));

        // If server is incommunicado...
        if (!WinDdePostMsg(pObjIDA->hwndServer, hwnd, WM_DDE_POKE,
                           pDdeStruct, pObjIDA->uflags))

          // ...delete packet.
          DosFreeMem(pDdeStruct);
      }
      break;

    //
    // Operator has expressed an interest in sounding an alarm.
    case UM_EXECUTE:
      {
        // The server recognizes 4 commands of the form
        // "ALARMn,xxxxxx" where 'n' is a number from 1 to 4
        // and 'xxxxxx' is a number from 000000 thru 999999.
        // Each "ALARMn" has a distinctive sound and the number
        // 'xxxxxx' specifies how long the alarm will sound.
        // No spaces are allowed anywhere within the command
        // string. 'xxxxxx' doesn't have to be zero padded.

        USHORT usAlarmNumber = SHORT1FROMMP(mp1);
        USHORT usAlarmDuration = SHORT2FROMMP(mp1);
        CHAR szData[80];

        sprintf(szData, "%s%d,%d",
                szItemName[ALARM],
                usAlarmNumber,
                usAlarmDuration);

        pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

        pDdeStruct = createDataPacket(pObjIDA->hwndServer,
                                      (PSZ)szItemName[ALARM],
                                      pObjIDA->pflags, (PSZ)szData,
                                      USHORT(strlen(szData)+1));


        // If server can't hear us, neither server (if acknowledge-
        // request is off) nor application's WM_DDE_ACK processing
        // (acknowledge-request is on) is going to release the data
        // packet so we'll do it here.
        if (!WinDdePostMsg(pObjIDA->hwndServer, hwnd,
                           WM_DDE_EXECUTE,
                           pDdeStruct, pObjIDA->uflags))
          DosFreeMem(pDdeStruct);
      }
      break;

    //
    // Client window has requested that we terminate the DDE link.
    case UM_TERM:

      // Get server handle from object window's IDA and...
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

      pDdeStruct = createDataPacket(pObjIDA->hwndServer,
                                    (PSZ)szItemName[TIME]);

      // Unlike other messages, a WM_DDE_TERMINATE data packet is
      // deleted by the application's WM_DDE_TERMINATE processing
      // rather than by application's WM_DDE_ACK processing (when
      // request-acknowledgement is enabled) or by the server (when
      // request-acknowledgement is disabled). However, it's still
      // true that if the server can't hear us, the packet isn't
      // going to get deleted unless we do it here.
      if (!WinDdePostMsg(pObjIDA->hwndServer, hwnd, WM_DDE_TERMINATE,
                         pDdeStruct, pObjIDA->uflags))
        DosFreeMem(pDdeStruct);
      break;

    //
    // Data received.
    //  mp1 = window handle of sender
    //  mp2 = DDESTRUCT data packet pointer
    //  DDESTRUCT->fsStatus:
    //     DDE_FACREQ =    If set, client receiver should return
    //                     an WM_DDE_ACK message
    //     DDE_FRESPONSE = If set, data is a reponse to a
    //                     WM_DDE_REQUEST else it is a response
    //                     to a WM_DDE_ADVISE
    case WM_DDE_DATA:
      {
        USHORT usMessage;

        // Get server's handle, data packet, object windows IDA,
        // client handle and client IDA.
        HWND hwndServer = HWNDFROMMP(mp1);
        pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);
        pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
        pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(pObjIDA->hwndClient,
                                                   QWL_USER);
        // if data was sent by our server...
        if (pObjIDA->hwndServer == hwndServer) {

          // If is data in this packet...
          if (pDdeStruct->offabData-pDdeStruct->offszItemName) {

            // ...move time into client IDA...
            strcpy((PCHAR)pClientIDA->szTime,
                   (PCHAR)DDES_PABDATA(pDdeStruct));

            // Force time update in client window (done by WM_PAINT
            // processing).
            WinInvalidateRect(pObjIDA->hwndClient, NULL, FALSE);

            // If acknowledge-data-processed enabled (i.e.,
            // continuous data update is operating in "warm" link
            // mode), acknowledge data receipt so server can send
            // us next data update.
            if (pObjIDA->pflags&DDE_FACKREQ) {

              // Say this is a FACK pack, that we weren't too busy
              // to process data and...
              pDdeStruct->fsStatus = USHORT((pDdeStruct->fsStatus|
                                             DDE_FACK)|~DDE_FBUSY);

              // ... setup for an acknowledment.
              usMessage = WM_DDE_ACK;
            }
          }
          else {

            // If there was no data in the packet the FNODATA
            // protocol must be in effect so dump the packet,
            // create a new one and...
            DosFreeMem(pDdeStruct);
            pDdeStruct = createDataPacket(pObjIDA->hwndServer,
                                          (PSZ)szItemName[TIME]);

            // ...setup to get the data via a one-shot update.
            usMessage = WM_DDE_REQUEST;
          }

          // If server isn't listening, tell client window and
          // let it decide how to deal with it.
          if (!WinDdePostMsg(pObjIDA->hwndServer, hwnd, usMessage,
                             pDdeStruct, pObjIDA->uflags)) {

            DosFreeMem(pDdeStruct);

            // Message which failed to ship is passed in mp1.
            WinPostMsg(pObjIDA->hwndClient, UM_NOSERVERRESPONSE,
                       MPFROMSHORT(WM_DDE_ACK), 0L);
          }

          // Idle is advise mode.
          pObjIDA->usRequest = 0;
        }
      }
      break;

    //
    // Acknowledge receipt of data
    //   mp1 = sender's window handle
    //   mp2 = pointer to DDESTRUCT data packet
    //     DDESTRUCT usStatus:
    //       DDE_FACK:         1 = accepted, 0 = rejected
    //       DDE_FBUSY:        0 = not busy, 1 = busy
    //       DDE_FAPPSTATUS:   1 = not understood, ignored
    //
    case WM_DDE_ACK:

      // In this case the server acknowledeges receipt of a message
      // from the application. This happens for all messages
      // sent by application to the server (e.g., WM_DDE_POKE,
      // WM_DDE_EXECUTE) irrespective of whether the acknowledge-
      // request option is enabled, excepting the initialization,
      // termination and acknowledgment messages.

      pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);
      if (!pDdeStruct->fsStatus&DDE_FACK)
        WinAlarm(HWND_DESKTOP, WA_ERROR);           // weak!
      if (pDdeStruct->fsStatus&DDE_FBUSY)
        WinAlarm(HWND_DESKTOP, WA_ERROR);           // weak!

      WinAlarm(HWND_DESKTOP, WA_NOTE);
      DosFreeMem(pDdeStruct);
      break;

    //
    // Request to disconnect the DDE server link.
    //   mp1 = sender's window handle
    //   mp2 = reserved
    case WM_DDE_TERMINATE:

      // Release data packet
      pDdeStruct = (PDDESTRUCT)PVOIDFROMMP(mp2);
      DosFreeMem(pDdeStruct);

      // ...and tell client window what we have done (client
      // window proc actually sends the WM_DDE_TERMINATE message
      // to the server.
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      WinPostMsg(pObjIDA->hwndClient, UM_TERM, 0L, 0L);

      // Guicide.
      WinDestroyWindow(hwnd);
      break;

    //
    case WM_TIMER:
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);
      hab = WinQueryAnchorBlock(hwnd);

      // If passing through here on a link initialization we
      // have spent 2 seconds waiting for a server response
      // and that's too long...
      if (LONGFROMMP(mp1) == TID_INITIATE) {
        if (pObjIDA->usInitiate) {

          // Stop the initialization timeout timer, give the
          // operator the bad news and release the object
          // window.
          WinStopTimer(hab, hwnd, TID_INITIATE);
          WinPostMsg(pObjIDA->hwndClient, UM_LINKINITFAILURE, 0L, 0L);
          WinDestroyWindow(hwnd);
        }
      }

      // If passing through here during a "cold" DDE link and
      // we're still waiting for an update, its altogether likely
      // that the server has gone south so...
      if (LONGFROMMP(mp1) == TID_REQUEST) {
        if (pObjIDA->usRequest) {

          // ...stop the request timer, and let the client window
          // decide how it wants to deal with the situation.
          WinStopTimer(hab, hwnd, TID_REQUEST);
          WinPostMsg(pObjIDA->hwndClient, UM_NOSERVERRESPONSE,
                     MPFROMSHORT(UM_DDE_TIMEOUT), 0L);
          pObjIDA->usRequest = 0;
        }
      }
      break;

    //
    //  Execution of command(s) requested
    //  mp1 = sender's window handle
    //  mp2 = DDESTRUCT pointer
    //    offabData = command(s) to be executed
    case WM_DDE_EXECUTE:
      // This application does nothing on receipt of WM_DDE_EXECUTE.
      break;

    //
    // Unsolicited data received.
    //   mp1 = sender's window handle
    //   mp2 = pointer to DDESTRUCT instance
    // Receiver is expected to reply with a positive WM_DDE_ACK
    // message (DDE_FACK bit = 1) if it accepts the unsolicited
    // data, or with a negative WM_DDE_ACK (DDE_FACK bit = 0)
    // if it does not.
    case WM_DDE_POKE:
      // This application does nothing on receipt of WM_DDE_POKE.
      break;


    //
    // Termination of ADVISE data transfer requested.
    // The DDESTRUCT data packet may specify the item name of the
    // data transfer that is to be discontinued. If it doesn't
    // (offszItemName is the offset to a zero length string) the
    // sender is requesting that we terminate all item links being
    // supported by this server.
    //   mp1 = sender's window handle
    //   mp2 = DDESTRUCT instance pointer
    case WM_DDE_UNADVISE:
      // This application does nothing on receipt of WM_DDE_UNADVISE.
      break;

    //
    // Continuous data update requested.
    //   mp1 = sender's window handle
    //   mp2 = pointer to DDESTRUCT instance
    case WM_DDE_ADVISE:
      // This application does nothing on receipt of WM_DDE_ADVISE.
      break;

    //
    // Data update requested.
    //   mp1 = server's window handle
    //   mp2 = pointer to DDESTRUCT instance
    //      offszItemName = item name to get
    //      usFormat      = DDE format in which data is to be
    //                      returned.
    case WM_DDE_REQUEST:
      // This application does nothing on receipt of WM_DDE_REQUEST.
      break;

    //
    // No matter how we exit the appllication we want any links
    // shut down.
    case WM_DESTROY:
      pObjIDA = (POBJIDA)WinQueryWindowPtr(hwnd, QWL_USER);

      WinDdePostMsg(pObjIDA->hwndServer, hwnd, WM_DDE_TERMINATE,
                    0L, pObjIDA->uflags);
      break;

    //
    default:
      mr = WinDefWindowProc(hwnd, msg, mp1, mp2);
      break;
  }

  return mr;
} // end of objWndProc()


/*-------------------------------------------------------------------
   setDDE_Controls - set/reset DDE menu controls. This is a client
                     window specific procedure.

   PARAMETERS:
     HWND hwnd - handle of the window which owns the DDE control
                 menu
     BOOL state - FALSE resets the DDE control options, i.e., the
                  'Initiate' DDE link item is enabled and the rest
                  are disabled. TRUE sets the DDE control options,
                  i.e., 'Initiate' is diabled and the rest are
                  enabled. If this parameter is no provided the
                  DDE controls are reset.

   RETURNS:   nada
*/

VOID setDDE_Controls(HWND hwnd, BOOL state) {
  enableMenuItem(hwnd, IDM_INIT, !state);
  enableMenuItem(hwnd, IDM_TERM, state);
  enableMenuItem(hwnd, IDM_ONE, state);
  enableMenuItem(hwnd, IDM_POKE, state);
  enableMenuItem(hwnd, IDM_START, state);
  enableMenuItem(hwnd, IDM_STOP, state);
} // end of setDDE_Controls()

//-------------------------------------------------------------------
VOID setTitle(HWND hwnd) {

  HWND hwndFrame = WinQueryWindow(hwnd, QW_PARENT);

  PCLIENTIDA pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwnd,
                                                        QWL_USER);

  // ...update titlebar status to reflect this update.
  sprintf(pClientIDA->szStatus, "(%p) > %s: %s [%s %s %s %s]",
          pClientIDA->hwndObj,
          pClientIDA->szLink,
          pClientIDA->szMsgType,
          pClientIDA->szNOFREE,
          pClientIDA->szRETRY,
          pClientIDA->szFACKREQ,
          pClientIDA->szFNODATA);
  WinSetWindowText(hwndFrame, (PSZ)pClientIDA->szStatus);
} // end of setTitle()


/*-------------------------------------------------------------------
   toggleCheck - checks menu option 'itemID' if it is unchecked,
                 unchecks it if it is checked.

   PARAMETERS:
     HWND  hwndMenu - topmost menu handle
     ULONG itemID   - menu item id

   RETURNS:  TRUE if item is checked, FALSE if item is not checked
             after toggle operation is performed.
*/

BOOL toggleCheck(HWND hwndMenu, ULONG itemID) {

  // Get check state of menu option,...
  USHORT bChecked = (USHORT)WinSendMsg(hwndMenu, MM_QUERYITEMATTR,
                                       MPFROM2SHORT(itemID, TRUE),
                                       MPFROMSHORT(MIA_CHECKED));

  bChecked = USHORT(bChecked?1:0);

  // ...try to toggle it. If attempt failed...
  if (!WinSendMsg(hwndMenu, MM_SETITEMATTR,
                  MPFROM2SHORT(itemID, TRUE),
                  MPFROM2SHORT(MIA_CHECKED, bChecked?~MIA_CHECKED:
                                                      MIA_CHECKED))){
    // ...do an alarm and report same state else...
    return bChecked;
  }

  // ...return toggle of previous check state.
  return !bChecked;
} // end of toggleCheck()

/*-------------------------------------------------------------------
   enableMenuItem - enable/disable an menu item 'usMenuItem', based
                    on the value of the 'bStatus' . It determines
                    the menu's window handle and then sends
                    WM_SETITEMATTR to change the item.
*/

VOID enableMenuItem(HWND hwndClient, USHORT usMenuItem, BOOL bStatus) {
  HWND hwndClientFrame;
  HWND hwndClientMenu;

  hwndClientFrame = WinQueryWindow(hwndClient, QW_PARENT);
  hwndClientMenu = WinWindowFromID(hwndClientFrame, FID_MENU);
  WinSendMsg(hwndClientMenu, MM_SETITEMATTR,
             MPFROM2SHORT(usMenuItem, TRUE),
             MPFROM2SHORT(MIA_DISABLED,
                          bStatus ? ~MIA_DISABLED : MIA_DISABLED));
} // end of enableMenuItem()


/*-------------------------------------------------------------------
   createDataPacket - allocates shared memory for a DDESTRUCT, its
                      item name and data, initializes the elements of
                      the allocation and arranges for that allocation
                      to be accesible to the both the DDE server and
                      sender processes.

   PARAMETERS:  hwndServer -  DDE server window handle
                pszItemName - item name
                usStatus    - DDE exchange status (default = 0)
                pszData    -  data (default = NULL)
                usDataLen  -  length of data (default = 0)
                usFormat    - data format (default = DDEFMT_TEXT)

   RETURNS: pointer to an initialized DDESTRUCT instance.
*/

PDDESTRUCT EXPENTRY createDataPacket(HWND hwndServer,
                                    PSZ pszItemName,
                                    USHORT fsStatus,
                                    PSZ pszData,
                                    USHORT usDataLen,
                                    USHORT usFormat) {

  // Calculate length of item name and, using this, the length
  // of the data passed to us and the size of a bare bones
  // DDESTRUCT instance, calculate the size of the complete
  // DDESTRUCT (i.e., with item name and item data) instance.
  INT usItemLen = USHORT(strlen((PCHAR)pszItemName)+1);
  INT usTotalLen = USHORT(sizeof(DDESTRUCT)+usItemLen+usDataLen);

  // Now allocate a DDESTRUCT of the appropriate dimensions
  // which can be shared (OBJ_GIVEABLE) and...
  PDDESTRUCT pDdeStruct;
  DosAllocSharedMem((PPVOID)&pDdeStruct, NULL, usTotalLen,
                    OBJ_GIVEABLE | PAG_COMMIT | PAG_WRITE);

  // ...load it with the appropriate data.
  pDdeStruct->cbData = usTotalLen;
  pDdeStruct->fsStatus = fsStatus;
  pDdeStruct->usFormat = usFormat;

  // The item name starts immediately after the last byte of
  // the DDESTRUCT instance proper and the data starts immediately
  // after the last byte of the item name.
  pDdeStruct->offszItemName = sizeof(DDESTRUCT);
  pDdeStruct->offabData = USHORT(sizeof(DDESTRUCT) + usItemLen);

  // Since neither the item name nor the data are ASCIIZ, we
  // use memcpy() to put them in place.
  memcpy(DDES_PSZITEMNAME(pDdeStruct), pszItemName, usItemLen);
  memcpy(DDES_PABDATA(pDdeStruct), pszData, usDataLen);

  // Finally, we give the server process access to the DDESTRUCT
  // instance...
  PID pidServer;
  TID tidServer;
  WinQueryWindowProcess(hwndServer, &pidServer, &tidServer);
  DosGiveSharedMem(pDdeStruct, pidServer, PAG_WRITE);

  // ...and return the DDESTRUCT allocation pointer.
  return pDdeStruct;
} // end of createDataPacket)


//-------------------------------------------------------------------
MRESULT EXPENTRY intervalDlgProc(HWND hwnd, ULONG msg,
                                 MPARAM mp1, MPARAM mp2) {
  PCLIENTIDA pClientIDA;
//HWND hwndClient;

  switch(msg) {
    case WM_INITDLG:
//       setSysMenu(hwnd);
//    hwndClient = WinQueryWindow(
//         WinWindowFromID(HWND_DESKTOP, DLG_INTERVAL), QW_OWNER);
      pClientIDA = (PCLIENTIDA)WinQueryWindowPtr(hwndClient, QWL_USER);
      WinSetDlgItemText(hwnd, DLGEF_INTERVAL,
                        (PSZ)pClientIDA->szInterval);
      return FALSE;

    case WM_COMMAND:
      switch(SHORT1FROMMP(mp1)) {

        // Operator canceled interval box changes.
        case DID_CANCEL:
          WinDismissDlg(hwnd, TRUE);
          break;

        // Operator wants interval changes effected.
        case DID_OK:

//        HWND hwndClient = WinQueryWindow(
//           WinWindowFromID(HWND_DESKTOP, DLG_INTERVAL), QW_OWNER);
          pClientIDA = (PCLIENTIDA)
             WinQueryWindowPtr(hwndClient, QWL_USER);
          WinQueryDlgItemText(hwnd, DLGEF_INTERVAL, 10,
                              (PSZ)pClientIDA->szInterval);
          WinDismissDlg(hwnd, TRUE);
          break;
       }
       break;

    default:
      return(WinDefDlgProc(hwnd, msg, mp1, mp2));
  }

  return FALSE;
} // end of intervalDlgProc()

